#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <pthread.h>
#include <netinet/in.h>

#include "context.h"
#include "adl.h"
#include "updateTree.h"
#include "downloads.h"
#include "source_get.h"
#include "leech_utils.h"
#include "leech_get.h"
#include "leech_protocol.h"
#include "utils.h"
#include "persist_get.h"
#include "backoff.h"

/* adaptive Download Algorithm, high level:
 * 	Variables: 
 * 		timer: time between actions
 * 		B_max: estimated max bandwidth
 * 		B_act: measured current instant throughput
 * 		B_old: measured previous instant throughput
 * 		B_avg: measured average throughput
 * 		B_std: measured standard deviation on B_avg
 * 		B_inc: amount to increment B_max estimate when we're wrong
 * 		T_inc: amount to backoff when we find we're to do nothing
 *
 * 	do one round /timer 
 * 		if( B_act > B_max)	// sanity check our bandwidth est
 * 			B_max = B_act + B_inc
 * 		if( B_avg > B_act - B_std) // throttle back
 * 			drop slowest connection
 * 			decrease update rate	
 * 			next
 * 		if( B_max > B_act + B_std) // increase throttle
 * 			add connection
 * 			increase update rate
 * 			next
 * 		else timer = timer + t_inc   // do nothing, stop bothering me :)
 */ 		

/********************************************************/

inline int adaptiveDoneYet(context * ctx){
	return blocksFull(ctx->dlBitMap);
}

/********************************************************/
/* decide whether we should go to phase2 of downloading (download
 * 	everything from anyone
 */

int phase2Check(context * ctx){
	if(ctx->phase2)
		return 0;
	if((blocksCountBits(ctx->dlBitMap)+blocksCountBits(ctx->ac->inProgress))
			>=ctx->numberOfBlocks){
		ctx->phase2=1;
		Dprintf("Going into phase2!!\n");
	}
	return 0;

}

/********************************************************/
int adaptiveDownloadFile(struct context *ctx){	// jumping off point
	struct timespec timeout;
	struct timeval now;
	int err=0;
	adaptiveContext * ac;
	ac = adaptiveContextNew(ctx);
	sleep(1);
	spawnAdaptiveDownloaderThread(ctx);
	pthread_mutex_lock(ac->lock);
	do {
		if(err>=0)
			pthread_mutex_unlock(ac->lock);
		err=0;
		Dprintf("\t------- MAX=%lf, ACT=%lf, OLD=%lf, AVG=%lf, STD=%lf\n, timer=%d\n",
				ac->max, ac->act, ac->old, ac->avg, ac->std, ac->timer);
		// update info
		phase2Check(ctx);	// do we go to phase2 or not?
		updateBandwidthEstimates(ctx);
		if( ac->max < ac->act)	// redo upper bound
			ac->max = ac->act + ac->b_inc;
		if((ac->max > (ac->act + ac->std)) || (ac->nConnections < 1)){	// increase throttle
			err=addConnection(ctx);
			increaseUpdateRate(ctx);
			if(err<0){	// err<0 --> retry
				Dprintf("adaptiveDownloadFile:: tried to add a connection, but couldn't\n");
				continue;
			}
			//if(err==1){	// starving!
			//	ctx->desiredNumOfNeighbors++;
			//}

		}
		else if( ac->avg > (ac->act + ac->std)){	// throttle back
			dropConnection(ctx);
			decreaseUpdateRate(ctx);
		}
		else {
			ac->timer += ac->timer_inc;
			Dprintf("adaptiveDL: chilling\n");
		}


		// sleep for a bit
		gettimeofday(&now,NULL);
		timeout.tv_sec = now.tv_sec + ac->timer;
		timeout.tv_nsec = now.tv_usec * 1000;
		pthread_mutex_lock(ac->lock);	// will get unlocked with we wait
		Dprintf2(2,"\t------- AC scheduler going to sleep\n");
		pthread_cond_timedwait(ac->cond,ac->lock,&timeout);
	}while(!adaptiveDoneYet(ctx));
	pthread_mutex_unlock(ac->lock);	// will get unlocked with we wait
	Dprintf("--- adaptive scheduler thread Exiting\n");
	return 0;
}
/**************************************************************/
int dprintBlockDone(context * ctx, int block){
	time_t now;
	if(DEBUGLEVEL==0)
		return 0;
	//now=time(NULL);
	now=0;
	Dprintf("BLOCKDONE %d	%ld %ld\n",block, now, now- ctx->startTime);
	return 0;
}

/**************************************************************/
adaptiveContext * adaptiveContextNew(context * ctx){
	adaptiveContext * ac;

	ac = malloc(sizeof(adaptiveContext));
	assert(ac);
	ac->max=ctx->adl_maxBandwidth;
	ac->timer=ctx->adl_timer;
	ac->maxConnections=ctx->adl_maxConnections;
	ac->timer_inc = ctx->adl_timer_inc;
	ac->b_inc = ctx->adl_bandwidth_inc;
	ac->act=ac->old=0;
	ac->nBandwidthEstimates =0;
	ac->bandwidthEstimates = malloc(sizeof(long)*MOVING_AVERAGE_LEN);
	ac->indexBandwidthEstimates=0;

	ac->nConnections=0;
	ac->connection = calloc(ac->maxConnections,sizeof(downloadContext *));

	ac->lock = malloc(sizeof(pthread_mutex_t));
	pthread_mutex_init(ac->lock,NULL);
	ac->cond = malloc(sizeof(pthread_cond_t));
	pthread_cond_init(ac->cond,NULL);

	ac->inProgress = blocksDuplicate(ctx->dlBitMap);	// blocks that are in progress
	blocksClear(ac->inProgress);

	ctx->ac=ac;
	ac->avg=ac->std=0;
	return ac;
}

/**************************************************************/
int updateBandwidthEstimates(context * ctx){
	double bps;
	double std;
	int i;
	adaptiveContext * ac= ctx->ac;

	Dprintf2(15,"enter updateBandwidthEstimates()\n");
	pthread_mutex_lock(ac->lock);
	bps=0;
	// calc current bandwidth
	for(i=0;i<ac->nConnections;i++){
		bps+= getDownloadBps(ac->connection[i]);
	}
	ac->old=ac->act;
	ac->act=bps;
	// short cut if no data
	if(ac->nConnections==0){
		ac->avg=ac->std = 0;
		pthread_mutex_unlock(ac->lock);
		Dprintf2(15,"leave updateBandwidthEstimates(): path1\n");
		return 0;
	}
	// add new bandwidth estimate
	ac->bandwidthEstimates[ac->indexBandwidthEstimates]=bps;
	ac->indexBandwidthEstimates = (ac->indexBandwidthEstimates + 1)%
		MOVING_AVERAGE_LEN;
	if(ac->nBandwidthEstimates<MOVING_AVERAGE_LEN)
		ac->nBandwidthEstimates++;
	// update average HERE
	ac->avg=0;
	for(i=0;i<ac->nBandwidthEstimates;i++)
		ac->avg += ac->bandwidthEstimates[i];
	ac->avg = ac->avg/ (double) ac->nBandwidthEstimates;	// new average
	std=0;
	for(i=0;i<ac->nBandwidthEstimates;i++)
		std += (ac->bandwidthEstimates[i]-ac->avg)*
			(ac->bandwidthEstimates[i]-ac->avg);
	ac->std = sqrt(std);				// new standard deviation
	pthread_mutex_unlock(ac->lock);
	Dprintf2(15,"leave updateBandwidthEstimates(): path2\n");
	return 0;
}

/********************************************************/
// assumes ac and connection is locked!
int reassignFreeDownloader(context * ctx,int i){
	int block;
	int err=0;
	int shouldBackoff;
	adaptiveContext * ac=ctx->ac;
	shouldBackoff=0;
	if(ac->connection[i]->source)
		//shouldBackoff=backoff_reassign_bias(ctx);
		shouldBackoff=backoff(ctx);
	Dprintf2(5,"reassignFreeDownloader:: shouldbackOff==%d; source ==%d\n",shouldBackoff,ac->connection[i]->source);
	if(shouldBackoff)
		block=-1;
	else
		block = getNewAssignment(ctx,ac->connection[i]);
	Dprintf2(5,"reassignFreeDownloader:: shouldbackOff==%d; block ==%d\n",shouldBackoff,block);
	if(block == -1 ){ // no new assignment possible
		if(adaptiveDoneYet(ctx)){	// no point in reassigning; we're done
			Dprintf("reassignFreeDownloader: short cutting out; DONE\n");
			return 0;
		}
		if(shouldBackoff){		// don't reassign back'ed off source connection
			Dprintf("reassignFreeDownloader: calling killConnection b.c of BACKOFF\n");
			killConnection(ctx,i);
			return 1;		
		}
		Dprintf("reassignFreeDownloaders:: tried to reassign %s connection:"
				" Failed; conn #%d\n", ac->connection[i]->source?"source":"peer",i);
		updateTreeMarkLeafNotBusy(ctx, ac->connection[i]->conn);
		killConnection(ctx,i);
		pthread_cond_signal(ac->cond);	// signal scheduler to find new people
		return 1;
	} else{
		ac->connection[i]->current=0;
		ac->connection[i]->block=block;
		ac->connection[i]->offset=ac->connection[i]->dbuflen=0;	// re-init
		blocksSet(ac->inProgress,block);	// set ATOMIC note in adl.h file
		Dprintf("reassignFreeDownloaders:: Reassign %s connection to"
			       " block %d; conn #%d\n", ac->connection[i]->source?"source":"peer",
			       block,i);
		if(ac->connection[i]->source)
			err=persist_source_download(ctx,ac->connection[i]);
		else
			err=persist_peer_download(ctx,ac->connection[i]);
		if(err){
			Dprintf("reassignFreeDownloader:: got error reassigning %s"
					" connection; killing conn #%d\n", 
					ac->connection[i]->source?"source":"peer",i);
			pthread_cond_signal(ac->cond);	// signal scheduler to find new people
			killConnection(ctx,i);
		}
	}
	return 0;
}

/********************************************************/

void * adaptiveDownloaderThread(void * ctxP){
	context * ctx= (context *) ctxP;
	adaptiveContext * ac = ctx->ac;
	fd_set rds;
	int i,max,err,count,len;
	//size_t offset;
	struct timeval tv;
	downloadContext *dlc;
	//unsigned char dbuf[DBUFSIZE];
	//int bufpos = 0;
	int started=0;

	while(!adaptiveDoneYet(ctx)){
		tv.tv_sec=ADL_DEFAULT_SELECT_TIME;
		tv.tv_usec=0;
		FD_ZERO(&rds);
		max=0;
		Dprintf2(15,"adaptiveDownloaderThread:: TOP of loop\n");
		// build up FD set
		pthread_mutex_lock(ac->lock);
		if(ac->nConnections<1){
			if(started&& !ctx->backoffDidBackOff){
				Dprintf("Badness: NO CONNECTIONS! Trying to add some again\n");
				pthread_cond_signal(ac->cond);	// signal scheduler to find new people
			}
			pthread_cond_wait(ac->cond,ac->lock);	// wait for a new connection
			pthread_mutex_unlock(ac->lock);
			continue;
		}
		started=1;
		for(i=0;i<ac->nConnections;i++){
			FD_SET(ac->connection[i]->sock,&rds);
			max = MAX(max,ac->connection[i]->sock);
		}
		pthread_mutex_unlock(ac->lock);
		// do select()
		err=select(max+1,&rds,NULL,NULL,&tv);
		if(err==0)
			continue;	// didn't read anything
		if(err<0){
			perror("adaptiveDownloaderThread::select");
			continue;
		}
		// do reads from each one
		pthread_mutex_lock(ac->lock);
		// FIXME: add timeouts for connections not serving anything
		for(i=0;i<ac->nConnections;i++){
			if(FD_ISSET(ac->connection[i]->sock,&rds)){
				dlc=ac->connection[i];	// less typing
				pthread_mutex_lock(ac->connection[i]->lock);
				count=read(ac->connection[i]->sock,&dlc->dbuf[dlc->dbuflen],
					       DBUFSIZE- dlc->dbuflen);
				Dprintf2(15,"adaptiveDownloaderThread:: \tDONE read on conn %d\n",i);
				if(count<=0){	// there was an error reading
					pthread_mutex_unlock(ac->connection[i]->lock);
					Dprintf("ADL: Error with connection: killing it\n");
					if(!ac->connection[i]->source)	// remove him from list
						updateTreeRevoke(ctx->updateTree, 
								connection_get_peer_ip(ac->connection[i]->conn),
								connection_get_peer_port(ac->connection[i]->conn)) ;
					blocksClearBit(ac->inProgress, ac->connection[i]->block);
					killConnection(ctx,i);
					pthread_cond_signal(ac->cond);	// signal scheduler to find new people
					continue;
				}
				if(dlc->offset==0)	// mark where in the file to write
					dlc->offset=ctx->blockSize * dlc->block;
				dlc->dbuflen+=count;	// update the buffer pointer
				if(dlc->dbuflen>=DBUFSIZE){
					// take MIN, incase server spit too much out
					count= MIN(count, ctx->blockSize - dlc->current);	// doesn't happen 
					Dprintf2(15,"adaptiveDownloaderThread:: \t\tleech_write \n");
					leech_file_write(ctx,dlc->dbuf,dlc->offset, dlc->dbuflen);// write to file
					Dprintf2(15,"adaptiveDownloaderThread:: \t\tcheckpoint ONE\n");
					dlc->dbuflen=0;
					dlc->offset+=DBUFSIZE;
				}
				dlc->current +=count;
				dlc->bytes +=count;
				// calc the expected amount
				if(ac->connection[i]->block == (ctx->numberOfBlocks-1))  // last block
					len = ctx->fileSize - ((ctx->numberOfBlocks-1)*(ctx->blockSize));
				else
					len = ctx->blockSize;
				Dprintf2(15,"adaptiveDownloaderThread:: \t\tcheckpoint TWO\n");
				if(ac->connection[i]->current >= len){
					// mark this block as done
					Dprintf2(15," adaptiveDownloaderThread:: DONE with block; doing cleanup\n");
					if(dlc->dbuflen!=0){
						leech_file_write(ctx,dlc->dbuf,dlc->offset, dlc->dbuflen);// write to file
					}
					blocksSet(ctx->dlBitMap,ac->connection[i]->block);
					dprintBlockDone(ctx,ac->connection[i]->block);
					if(ac->connection[i]->source)
						ctx->sourceBlocksDld++;
					else
						ctx->peerBlocksDld++;
					//if(!ac->connection[i]->source)
					//  updateTreeMarkLeafNotBusy(ctx, ac->connection[i]->conn);
					blocksClearBit(ac->inProgress, ac->connection[i]->block);
					pthread_mutex_unlock(ac->connection[i]->lock);
					Dprintf2(15," adaptiveDownloaderThread:: calling reassignFreeDownloader\n");
					reassignFreeDownloader(ctx,i);	// reassign this connection
					Dprintf2(15," adaptiveDownloaderThread:: DONE reassignFreeDownloader\n");
					continue;
				}
				Dprintf2(15,"adaptiveDownloaderThread:: \t\tcheckpoint THREE\n");
				pthread_mutex_unlock(ac->connection[i]->lock);
			}
		}
		pthread_mutex_unlock(ac->lock);
	}
	Dprintf("--- adaptive downloader thread Exiting\n");
	return NULL;
}


/* int addConnection(struct context * ctx);
 * 	find someone in the world that we can get a new block from
 */


download * dl_list[DL_MAXNODESPERBLOCK];	// out here just to avoid putting them on the stack
download * dl_list1[DL_MAXNODESPERBLOCK];

int addConnection(struct context * ctx){
	int i,j,bstart,b=0,b1=0;
	adaptiveContext *ac = ctx->ac;
	int count,count1;
	int ret=0;

	bstart = abs(mrand48())% ctx->numberOfBlocks;	// pick a random block as 
						// a starting point, and search from there
	count=count1=0;
	// pick a random block
	for(i=0;i<ctx->numberOfBlocks;i++){
		b = (i+bstart)%ctx->numberOfBlocks;
		// if we already have it, continue
		if(blocksIsSet(ctx->dlBitMap,b))	
			continue;
		// if we are already working on it, continue
		// special case: if we are working on it as a source connection,
		// try to get it from a peer anyway
		if(ctx->phase2==0){
			pthread_mutex_lock(ac->lock);
			for(j=0;j<ac->nConnections;j++)
				if((ac->connection[j]->source!=1)&&(ac->connection[j]->block==b))
					break;
			pthread_mutex_unlock(ac->lock);
			if(j<ac->nConnections)	// did we break'ed early
				continue;
		}
		// if we can't find it, continue
		count= DL_MAXNODESPERBLOCK;
		updateTreeSearch(ctx->updateTree, b, &count, dl_list);
		//Dprintf("ADL: UpdateTreeSearch1 found %d peers\n", count);
		if(count!=0)
			break;	// found someone
	}
	if(ctx->pickTwo){
		// pick another random block
		bstart = abs(mrand48())% ctx->numberOfBlocks;	// pick a random block as 
		for(i=0;i<ctx->numberOfBlocks;i++){
			b1 = (i+bstart)%ctx->numberOfBlocks;
			// if we already have it, continue
			if(blocksIsSet(ctx->dlBitMap,b1))	
				continue;
			// if we are already working on it, continue
			pthread_mutex_lock(ac->lock);
			for(j=0;j<ac->nConnections;j++)
				if(ac->connection[j]->block==b1)
					break;
			pthread_mutex_unlock(ac->lock);
			if(j<ac->nConnections)	// did we break'ed early
				continue;
			// if we can't find it, continue
			count1= DL_MAXNODESPERBLOCK;
			updateTreeSearch(ctx->updateTree, b1, &count1, dl_list1);
			//Dprintf("ADL: UpdateTreeSearch12 found %d peers\n", count1);
			if(count1!=0)
				break;	// found someone
		}
		if(count1<count){	// if this block is less replicated
			Dprintf("PickTwo: Picked the other block\n");
			b=b1;		// get the less replicated block
			count=count1;
			for(i=0;i<count;i++)
				dl_list[i]=dl_list1[i];
		}
	}

	if(count>0) // did we find any suitable blocks?
		ret= addPeerConnection(ctx,b,count, dl_list);
	else if(!backoff(ctx))
		ret= addSourceConnection(ctx);
	/*		DON'T DO THIS; if we fail a peer connection; just try again later
	if((count>0)&&(ret!=0)&&(!backoff(ctx)))	// if we tried to add a peer connection, and failed, try a
		ret= addSourceConnection(ctx);  // source connection
		*/
	pthread_mutex_lock(ac->lock);
	pthread_cond_signal(ac->cond);	// signal to the downloader thread that something new has been added
	pthread_mutex_unlock(ac->lock);
	return ret;
}

/* int addSourceConnection(struct context * ctx);
 * 	add a connection to the original source, if one doesn't already exist
 */

int addSourceConnection(struct context * ctx){
	int i;
	adaptiveContext *ac=ctx->ac;
	downloadContext * dlc;
	int block;
	pthread_mutex_lock(ac->lock);
	for(i=0;i<ac->nConnections;i++)
		if(ac->connection[i]->source)
			break;
	pthread_mutex_unlock(ac->lock);
	if(i<ac->nConnections){
		Dprintf2(4,"adaptiveDL: tried to add but STARVING... \n");// want to add someone, can't add
		return 1; // ANYTHING
	}
	if(ac->nConnections>=ac->maxConnections){
		Dprintf("adaptiveDL: too many connections open\n");
		return 2;
	}
	block = updateTreeFindMissingBlock(ctx->updateTree,ctx);
	if(block == -1 ){
		if((ctx->phase2 == 0 )|| (ctx->backoffDidBackOff)){
			Dprintf("adaptiveDL: tried to add source; but should be adding peer\n");
			return 1;
		}
		for(block=0;block<ctx->numberOfBlocks;block++)
			if(!blocksIsSet(ctx->dlBitMap,block))
				break;
		if(block>=ctx->numberOfBlocks)	// must have finished everything
			return 1;
	}

	Dprintf("adaptiveDL: Adding Source Connection\n");
	dlc = persist_source_download_init(ctx);
	if(!dlc){
		fprintf(stderr,"addSourceConnection::Error connecting to source\n");
		return 2;
	}

	dlc->block = block;
	persist_source_download(ctx,dlc);
	pthread_mutex_lock(ac->lock);
	ac->connection[ac->nConnections++]=dlc;	// register new connection
	pthread_mutex_unlock(ac->lock);
	return 0;
}
	
/* int addPeerConnection(struct context * ctx);
 * 	add a connection to one of the passed peers; just pick one at random for now
 */
int addPeerConnection(struct context * ctx, int block, int count, download ** list){
	int i,start,target,err;
	downloadContext * dlc=NULL;
	adaptiveContext * ac = ctx->ac;

	Dprintf("adaptiveDL: adding a peer connection\n");
	if(ac->nConnections>=ac->maxConnections){
		Dprintf("adaptiveDL: too many connections open\n");
		return 2;
	}
	start = abs(mrand48())%count;
	for(i=0;i<count;i++){	// repeat, until we find one that accepts us
		target = (i+start)%count;
		dlc = persist_peer_download_init(ctx,list[target]);
		if(dlc)
			break;
		/* NOTE: if people aren't accepting connections on that port, we drop
		 * 	them in persist_peer_download_init(), not here
		 */
	}
	if(!dlc){
		Dprintf("addPeerConnection::Error connecting to all %d peers\n",count);
		return -3;
	}
	dlc->block=block;
	blocksSet(ac->inProgress,block);	 // see ATOMIC note in adl.h
	err=persist_peer_download(ctx,dlc);
	if(err){
		Dprintf("persist_peer_download() of block %d failed; err=%d\n",block,err);
		return -1;
	}
	pthread_mutex_lock(ac->lock);
	ac->connection[ac->nConnections++]=dlc;	// register new connection
	pthread_mutex_unlock(ac->lock);

	return 0;
}
/* int spawnAdaptiveDownloaderThread(struct context * ctx);
 * 	spawn off a thread that just calls adaptiveDownloaderThread
 */
int spawnAdaptiveDownloaderThread(struct context * ctx){
	int err;
	pthread_t * thread;
	thread = malloc(sizeof(pthread_t));
	err=pthread_create(thread,NULL,adaptiveDownloaderThread,ctx);
	if(err){
		perror("spawnAdaptiveDownloaderThread::pthread_create");
		return 1;
	}
	pthread_detach(*thread);
	return 0;
}

/* double getDownloadBps(downloadContext *dlc);
 *	return the BPS on this connection, reset the timers and stats
 */

double getDownloadBps(downloadContext *dlc){
	double bps;
	unsigned long now,t;
	assert(dlc);
	pthread_mutex_lock(dlc->lock);
	now = time(NULL);
	t = now - dlc->time_start;	// calc time delta
	if(t<1)			// prevent DIV/zero
		t=1;
	bps = dlc->bytes/(double)t;
	dlc->bytes=0;
	dlc->time_start=now;
	dlc->obps=bps;
	pthread_mutex_unlock(dlc->lock);
	return bps;
}

/* int dropConnection(struct context * ctx);
 * 	pick the slowest connection, and drop it
 */

int dropConnection(struct context * ctx){
	adaptiveContext *ac = ctx->ac;
	int i,min=0;
	double minbps=-1;
	if(ac->nConnections>1)
		Dprintf("adaptiveDL: Dropping a connection\n");
	else{
		Dprintf("adaptiveDL: Tried to drop only connection: BOGUS!\n");
		return 0;
	}
	pthread_mutex_lock(ac->lock);
	for(i=0;i<ac->nConnections;i++){
		if(minbps==-1){
			minbps=ac->connection[i]->obps;
			min=i;
		} else {
			minbps=MIN(ac->connection[i]->obps,minbps);
			min=i;
		}
	}

	assert(minbps!=-1);	// tried to drop a connection when we have none :(
	killConnection(ctx,min);
	pthread_mutex_unlock(ac->lock);
	return 0;
}

/* int killConnection(struct context * ctx,int i)	
 * 	actually shutdown and remove connection i
 * 	ASSUMES that ac is locked, and that the ith connection is not locked
 * 		(that would be silly)
 */

int killConnection(struct context * ctx,int i){
	adaptiveContext * ac= ctx->ac;
	assert(i<ac->nConnections);
	Dprintf("killing adaptive %s Connection\n", ac->connection[i]->source?"source":"peer");
	shutdown(ac->connection[i]->sock,SHUT_RDWR);
	close(ac->connection[i]->sock);
	if(ac->connection[i]->conn)
		connection_destroy(ac->connection[i]->conn);
	free(ac->connection[i]->lock);
	free(ac->connection[i]);
	ac->connection[i]=NULL;
	// remove it from the connection list
	ac->nConnections--;
	if(i==ac->nConnections) // this was at the end of the list
		return 0;
	// else replace it with the last connection
	ac->connection[i]=ac->connection[ac->nConnections];
	return 0;
}
/***********************************************************************/
// assumes ac is locked
int getNewAssignment(struct context * ctx,downloadContext *dlc){
	adaptiveContext *ac=ctx->ac;
	struct sockaddr_in sa;
	int err,len,block,bstart;
	int i,wrapped;
	unsigned short port;
	Dprintf2(3,"getNewAssignment::\n");
	if(dlc->source){
		block= updateTreeFindMissingBlock(ctx->updateTree,ctx);
		if(block !=-1){
			Dprintf2(3,"\tsource assigned %d\n",block);
			return block;
		}
		// we know about everything, just grab something we don't have
		for(i=0;i<ctx->numberOfBlocks;i++){
			if(!blocksIsSet(ctx->dlBitMap,i)){
				Dprintf2(3,"\tsource PHASE 2 assigned %d\n",i);
				return i;
			}
		}
		Dprintf("\tsource wants NOTHING\n");
		return -1; 	// there really is nothing else we need
	}

	len = sizeof(sa);
	err= getpeername(dlc->sock, (struct sockaddr *)&sa, &len);
	if(err){
		perror("getNewAssignment::getpeername");
		return -1;
	}
	bstart = abs(mrand48())%ctx->numberOfBlocks;	// pick random start point
	block=bstart;
	wrapped=0;
	port = ntohs(sa.sin_port);
	do {
		Dprintf2(4,"STUPID FUCKING LOOP(getNewAssignment): block = %d\n",block);
		// find a block that this guy has, that we don't starting at 'block'
		block = updateTreeFirstNeedBlock(ctx,sa.sin_addr.s_addr,port, block);
		if(block <0){
			if(wrapped)
				break; 	// nothing we can get from them, return
			else{
				wrapped=1;	// try from beginning
				block=0;
				continue;
			}
		}
		// check to see if we are already working on this block
		for(i=0;i<ac->nConnections;i++){
			if(ac->connection[i]->block == block)
				break;
		}
		if(i==ac->nConnections) // we found a block that no one is using
			break;
		block++;		// else next block
	} while((wrapped==0)||(block < bstart));
	
	Dprintf2(3,"\tPEER assigned %d\n",block);
	return block;
}


/* int increaseUpdateRate(struct context * ctx)
 * 	increase the rate of which people send us updates,
 * 	because we have the bandwith to spare
 */
int increaseUpdateRate(context * ctx){
	int old;
	old = ctx->desiredNumOfUpdates;
	ctx->desiredNumOfUpdates++;
	return old;
}
/* int decreaseUpdateRate(struct context * ctx);
 * 	decrease the rate at which people send us updates,
 * 	because we have exceeded our bandwidth rates
 * 	NOTE - this can effectively turn off all updates,
 * 		which is ok, b/c in theory if we know where
 * 		everything is, we don't need to waste bandwidth findout
 * 		out about duplicates (is this true?)
 */
int decreaseUpdateRate(context * ctx){
	int old;
	old = ctx->desiredNumOfUpdates;
	ctx->desiredNumOfUpdates/=2;
	if(ctx->desiredNumOfUpdates<0)
		ctx->desiredNumOfUpdates=0;	// never go below zero
	return old;
}

