#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <netinet/in.h>

#define BITS	8

#include "blocks.h"
#include "utils.h"

#define GETBIT(x,y)	((x)&(1<<(y)))
#define SETBIT(x,y)	(x=((x)|(1<<(y))))
#define CLEARBIT(x,y)	(x=((x)&(~(1<<(y)))))

/* blocksNew():
 * 	return a new blocks structure
 */
blocks * blocksNew(int blockSize, int totalSize){
	blocks *ret;
	int num;
	
	ret = malloc(sizeof(blocks));
	if(!ret){
		perror("blocksNew::malloc");
		return NULL;
	}
	assert(blockSize>0);
	assert(totalSize>0);
	ret->blockSize = blockSize;
	ret->totalSize=totalSize;
	ret->numBlocks = ceil((double)totalSize/blockSize);
	ret->refCount=1;
	num = ceil(((double)ret->numBlocks)/(double)BITS);
	assert(num>0);
	ret->bitMap = malloc(num);
	memset(ret->bitMap, 0, num);
	if(!ret->bitMap){
		perror("blocksNew::malloc");
		free(ret);
		return NULL;
	}


	// initialize lock
	ret->lock = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	if (ret->lock == NULL) {
	  free(ret);
	  return NULL;
	}
	pthread_mutex_init(ret->lock, NULL); 
	blocksClear(ret);	
	return ret;
}


/* int blocksIsSet(blocks * b,int block);
 * 	return zero if bit is not set
 * 	non-zero otherwise	 (i.e. True)
 */

inline int blocksIsSet(blocks * b,int block){
	int seg,off;	// names kinda misleading, but I like them :)
	int ret;
	assert(block>=0);
	assert(block<b->numBlocks);
	pthread_mutex_lock(b->lock);
	seg=block >> 3;
	off=block%BITS;
	ret = GETBIT(b->bitMap[seg],off);
	pthread_mutex_unlock(b->lock);
	return 	ret;
}
	
int blocksIsSetNoLock(blocks * b,int block){
	int seg,off;	// names kinda misleading, but I like them :)
	int ret;
	assert(block>=0);
	assert(block<b->numBlocks);
	seg=block >> 3;
	off=block%BITS;
	ret = GETBIT(b->bitMap[seg],off);
	return 	ret;
}
	

/* int blocksSet(blocks * b,int block);
 * 	set the bit coresponding to the block in the blocklist b
 */

int blocksSet(blocks * b,int block){
	int seg,off;	// names kinda misleading, but I like them :)
	int ret;
	assert(block>=0);
	assert(block<b->numBlocks);
	pthread_mutex_lock(b->lock);

	seg=block/BITS;
	off=block%BITS;
	ret = SETBIT(b->bitMap[seg],off);
	pthread_mutex_unlock(b->lock);
	return 	ret;
}

/* int blocksClear(blocks *b);
 * 	clears all of the bits in the bitmask
 */

int blocksClear(blocks *b){
	int i,seg,off;
	assert(b);
	pthread_mutex_lock(b->lock);
	for(i=0;i<b->numBlocks;i++){
		seg=i/BITS;
		off=i%BITS;
		CLEARBIT(b->bitMap[seg],off);
	}
	pthread_mutex_unlock(b->lock);
	return 0;
}

/* int blocksClearBit(blocks *b,int block);
 * 	clear a specific block
 */
int blocksClearBit(blocks *b,int block){
	int seg,off;
	assert(b);
	pthread_mutex_lock(b->lock);
	seg=block/BITS;
	off=block%BITS;
	CLEARBIT(b->bitMap[seg],off);
	pthread_mutex_unlock(b->lock);
	return 0;
}

/* void blocksFree(blocks *b);
 * 	free any blocks resources
 */

void blocksFree(blocks *b){
  int rv;
	assert(b);
	assert(b->bitMap);
	pthread_mutex_lock(b->lock);
	if((b->refCount--)>0){
		pthread_mutex_unlock(b->lock);
		return;
	}

	// destroy the lock
        rv = pthread_mutex_destroy(b->lock);
	// if rv == EBUSY somebody is locking the queue
	// assert(rv != EBUSY) somehow the EBUSY is not defined on my system
	assert(rv == 0);

	free(b->bitMap);
	free(b);
}


/* int blocksMerge(a,b):
 * 	return blocks a union of a and b
 *		assumes all blocks are allocated and of the same size
 */

blocks * blocksMerge(blocks * a, blocks *b){
	int i,seg, off;
	blocks *dst;

	assert(a);
	assert(b);
	assert(a->numBlocks== b->numBlocks);
	dst = blocksNew(a->blockSize,b->totalSize);
	seg=a->numBlocks/BITS;
	off=a->numBlocks%BITS;
	if(off!=0) seg++;
	for(i=0;i<seg;i++)
		dst->bitMap[i]=a->bitMap[i]|b->bitMap[i];

	return dst;
}



int blocksIsByteRangeSet(blocks *b, size_t offset, size_t len) 
{
  size_t startBlock = offset/(b->blockSize);
  size_t endBlock = (offset+len-1)/(b->blockSize);
  int i;

  if(offset%(b->blockSize)){
	  Dprintf("blocksIsByteRangeSet:: non-block aligned access\n");
  }
  if((len != b->blockSize)&&( startBlock == (b->numBlocks-1))){	
	  // i.e. if len is not right, and we aren't at the last block
	  Dprintf("blocksIsByteRangeSet:: non-block aligned length \n");
  }
  if((startBlock>=b->numBlocks)||(endBlock>=b->numBlocks))
	  return 0;	// just plain out of range
  pthread_mutex_lock(b->lock);
  for (i = startBlock; i <= endBlock; ++i) {
    if (!blocksIsSetNoLock(b,i)) {
      pthread_mutex_unlock(b->lock);
      return 0;
    }
  }

  pthread_mutex_unlock(b->lock);
  return 1;
}

/* int blocksRead(blocks * b,int fd,int timeout);
 *	 read the bitmap from the fd
 */

int blocksRead(blocks * b,int fd,int timeout){
	int count;
	int rem_count;
	int err,status;
	
	count = b->numBlocks/BITS;
	if(b->numBlocks%BITS)
		count++;

	err = timeout_readall(&status,fd, &rem_count, sizeof(rem_count),timeout);
	if (err) return err;
	rem_count = ntohl(rem_count);
	assert(rem_count == count);
	err = timeout_readall(&status,fd, b->bitMap, count,timeout);
	if(err) return err;
	if(status<count)
		return -1;

	/*
	   { int i;
	Dprintf("Read: ");
        for (i = 0; i < count; i++)
          fprintf(stderr, "%d ", b->bitMap[i]);
	  fprintf(stderr, "\n");
	  }
	  */
	return 0;
}

/* int blocksWrite(blocks * b,int fd,int timeout);
 * 	write the raw bitmap to the fd
 */

int blocksWrite(blocks * b,int fd,int timeout){
	int count;
	int send_count;
	int err,status;
	
	count = b->numBlocks/BITS;
	if(b->numBlocks%BITS)
		count++;
	send_count = htonl(count);
	err = timeout_writeall(&status,fd, &send_count, sizeof(send_count), timeout);
	if (err) return err;
	err = timeout_writeall(&status,fd, b->bitMap, count,timeout);
	if(err) return err;
	if(status<count)
		return -1;
	/*        Dprintf("Wrote: ");
        for (i = 0; i < count; i++)
          fprintf(stderr, "%d ", b->bitMap[i]);
	  fprintf(stderr, "\n"); */
	return 0;
}

/* int blocksCountBits(blocks *b);
 * 	count the number of bits set in a block
 */

int blocksCountBits(blocks *b){
	int count=0;
	int i;
	assert(b);
	for(i=0;i<b->numBlocks;i++){
		if(blocksIsSet(b,i))
			count++;
	}

	return count;
}

/* int blocksFull(blocks *b)
 * 	return 1 if all bits are set, 0 otherwise
 */

int blocksFull(blocks *b){
	int i;
	assert(b);
	for(i=0;i<(b->numBlocks>>3);i++)
		if(b->bitMap[i]!=0xff)
			return 0;
	if(b->numBlocks%BITS){
		pthread_mutex_lock(b->lock);
		for(i=(b->numBlocks>>3);i<b->numBlocks;i++)
			if(!blocksIsSetNoLock(b,i)){
				pthread_mutex_unlock(b->lock);
				return 0; 
			}
		pthread_mutex_unlock(b->lock);
	}
	return 1;
}

/* int blocksCmp(blocks * a, blocks * b);
 * 	return 1> of a>b, 1< if a<b, 0 otherwise
 */
int blocksCmp(blocks * a, blocks * b){
	int i;
	int diff;
	assert(b);
	assert(a);
	if(a->numBlocks!=b->numBlocks)
		return 1;
	for(i=0;i<(b->numBlocks>>3);i++)
		if((diff=a->bitMap[i]-b->bitMap[i]))
			return diff;
	if(b->numBlocks%BITS){
		for(i=(b->numBlocks>>3);i<b->numBlocks;i++)
			if((diff=blocksIsSet(a,i)-blocksIsSet(b,i)))
				return diff; 
	}
	return 0;
}




/* blocks * blocksDuplicate(blocks *b);
 * 	make a copy of this blocks struct and return it
 */
blocks * blocksDuplicate(blocks *b){
	blocks *ret;
	int len;
/*	ret = malloc(sizeof(blocks));
	if(!ret){
		perror("blocksDuplicate::malloc");
		abort();
	}
	ret->blockSize = b->blockSize;
	ret->numBlocks = b->numBlocks;
	ret->totalSize = b->totalSize;
	len = ret->numBlocks/BITS;
	if(ret->numBlocks%BITS)
		len++;
	ret->bitMap = malloc(len);
	if(!ret->bitMap){
		perror("blocksDuplicate::malloc");
		abort();
	}
	memcpy(ret->bitMap,b->bitMap,len);
	ret->lock = malloc(sizeof(pthread_mutex_t));
	if(!ret->lock){
		perror("blocksDuplicate::malloc");
		abort();
	}
	pthread_mutex_init(ret->lock,NULL);
	*/

	assert(b);
	ret = blocksNew(b->blockSize,b->totalSize);
	assert(ret);
	len = ret->numBlocks/BITS;
	if(ret->numBlocks%BITS)
		len++;
	memcpy(ret->bitMap,b->bitMap,len);

	return ret;
}

/* int blocksIncRef(blocks *b);
 * 	increment the reference counting on this struct
 */

int blocksIncRef(blocks *b){
	assert(b);
	pthread_mutex_lock(b->lock);
	b->refCount++;
	pthread_mutex_unlock(b->lock);
	return 0;
}





/* testing stuff
 */

int do_blocks_tests(){
	blocks *a,*b;
	int size = 1234234;
	int bs = 1024;
	int i;
	int num;

	printf("--- Doing blocks_tests()\n");

	num  = ceil((double)size/bs);
	assert(num>0);

	a = blocksNew(1024,size);
	b = blocksNew(1024,size);
	b=b;		// for now, ignore

	for(i=0;i<num;i++){
		assert(blocksIsSet(a,i) == 0);
		blocksSet(a,i);
		assert(blocksIsSet(a,i) != 0);
	}


	blocksFree(a);
	blocksFree(b);
	return 0;
}

