#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>

#include "utils.h"

#include "ftp_utils.h"

#define CRLF "\r\n"

int ftp_get_last_line(int theControlSock, 
		      char *theBuf, 
		      int theBufLen) {
  int err, status;

  err = timeout_read_line(&status, theControlSock, theBuf, theBufLen, FTP_TIMEOUT);
  if (err) {
    return err;
  }

  while ((theBuf[FTP_RETCODE_SIZE] != ' ') && (!err)) {
    err = timeout_read_line(&status, theControlSock, theBuf, theBufLen, FTP_TIMEOUT);
  }

  return 0;
}


int ftp_get_retcode(int theControlSock, 
		    char *tempBuf) {
  int err, result;

  err = ftp_get_last_line(theControlSock, tempBuf, FTP_MAX_LINE);
  if (err) {    // some problem with transport layer
    Dprintf("FTP SERVER: TCP error\n");
    result = FTP_TCP_ERR;
  }
  else {
    Dprintf("FTP SERVER: %s\n", tempBuf);
    result = atoi(tempBuf);
  }

  return result;
}

// will retrieve IP address and port number from string containing
// (n1,n2,n3,n4,p1,p2)
// IP address "n1.n2.n3.n4"
// port p1*256 + p2
void ftp_get_addr(const char *theBuf,
		  char* theIpOut,
		  unsigned int *thePortOut) {
  int i;
  char *ptr, *ptr2;

  ptr = strchr(&theBuf[FTP_RETCODE_SIZE], '(');
  assert(ptr);
  ++ptr;       // ptr points to the first character of n1
  // locate the n1,n2,n3,n4
  ptr2 = ptr;
  for (i = 0; i < 4; ++i) {
    ptr2 = strchr(ptr2, ',');
    assert(ptr2);
    ++ptr2;
  }    
  strncpy(theIpOut, ptr, ptr2-ptr-1); // ptr2 points to first char of p1
  theIpOut[ptr2-ptr-1] = '\0';
  // replace ',' by '.'
  ptr = theIpOut;
  for (i = 0; i < 3; ++i) {
    ptr = strchr(ptr, ',');
    *ptr = '.';
    ++ptr;
  }    

  // copying port number
  *thePortOut = atoi(ptr2);
  *thePortOut *= 256;
  ptr2 = strchr(ptr2, ',');
  assert(ptr2);
  ++ptr2;
  *thePortOut += atoi(ptr2);
}


int ftp_connect_cmd(int theControlSock) {
  char tempBuf[FTP_MAX_LINE];
  int serverRetcode;

  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_SERVER_READY) {
    // unexpected FTP retcode for this command
    return FTP_FTP_ERR;
  } 
  else {
    return FTP_CMD_OK;
  }
}


int ftp_login_cmd(int theControlSock,
const char* theUser,
const char* thePass) {
  char tempBuf[FTP_MAX_LINE];
  int serverRetcode;
  int err, count, status;

  // issue USER command
  count = snprintf(tempBuf, FTP_MAX_LINE, "USER %s%s", theUser, CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }
  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    return FTP_TCP_ERR;
  } 

  // issue PASS command
  count = snprintf(tempBuf, FTP_MAX_LINE, "PASS %s%s", thePass, CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }
  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_USER_LOGGED) {
    // unexpected FTP retcode for this command
    return FTP_FTP_ERR;
  } 
  else {
    return FTP_CMD_OK;
  }
}


int ftp_retr_cmd(int theControlSock,
		 char theType, 
		 const char *theFileName,
		 unsigned int theOffset,
		 int *theDataSockP) {
  char tempBuf[FTP_MAX_LINE];
  char ip[FTP_MAX_LINE];
  unsigned int port;
  int serverRetcode;
  int err, count, status;


  // issue PASV command
  // server will suply IP address and port number for the data connection
  count = snprintf(tempBuf, FTP_MAX_LINE, "PASV%s", CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }
  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_PASV_MODE_OK) {
    // unexpected FTP retcode for this command
    return FTP_FTP_ERR;
  } 
  else {
    // retrieve IP address and port number of data connection
    ftp_get_addr(tempBuf, ip, &port);
  }

  // issue TYPE command
  count = snprintf(tempBuf, FTP_MAX_LINE, "TYPE %c%s", theType, CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }
  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_TYPE_OK) {
    // unexpected FTP retcode for this command
    return FTP_FTP_ERR;
  } 

  // issue REST command
  count = snprintf(tempBuf, FTP_MAX_LINE, "REST %d%s", theOffset, CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }
  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_REST_OK) {
    // unexpected FTP retcode for this command
    return FTP_FTP_ERR;
  } 

  // issue RETR command
  count = snprintf(tempBuf, FTP_MAX_LINE, "RETR %s%s", theFileName, CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }

  // open TCP data connectio
  *theDataSockP = make_tcp_connection(ip, port);
  if (*theDataSockP < 0) {
    return FTP_TCP_ERR;
  }

  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    close(*theDataSockP);
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_FILE_REQ_OK) {
    // unexpected FTP retcode for this command
    close(*theDataSockP);
    return FTP_FTP_ERR;
  } 
  else {
    return FTP_CMD_OK;
  }
}

int ftp_list_cmd(int theControlSock,
		 const char* theFileGrpNameP, 
		 int *theDataSockP) {
  char tempBuf[FTP_MAX_LINE];
  char ip[FTP_MAX_LINE];
  unsigned int port;
  int serverRetcode;
  int err, count, status;


  // issue PASV command
  // server will suply IP address and port number for the data connection
  count = snprintf(tempBuf, FTP_MAX_LINE, "PASV%s", CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }
  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_PASV_MODE_OK) {
    // unexpected FTP retcode for this command
    return FTP_FTP_ERR;
  } 
  else {
    // retrieve IP address and port number of data connection
    ftp_get_addr(tempBuf, ip, &port);
  }

  // issue TYPE command
/*    count = snprintf(tempBuf, FTP_MAX_LINE, "TYPE A%s", CRLF); */
/*    Dprintf("FTP CLIENT: %s", tempBuf); */
/*    err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT); */
/*    if (err) { */
/*      return FTP_TCP_ERR; */
/*    } */
  // get server reply
/*    serverRetcode = ftp_get_retcode(theControlSock, tempBuf); */
/*    if (serverRetcode == FTP_TCP_ERR) { */
/*      return FTP_TCP_ERR; */
/*    }  */
/*    else if (serverRetcode != FTP_TYPE_OK) { */
/*      // unexpected FTP retcode for this command */
/*      return FTP_FTP_ERR; */
/*    }  */

  // issue LIST command
  count = snprintf(tempBuf, FTP_MAX_LINE, "LIST %s%s", theFileGrpNameP, CRLF);
  Dprintf("FTP CLIENT: %s", tempBuf);
  err = timeout_writeall(&status, theControlSock, tempBuf, count, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }

  // open TCP data connection
  *theDataSockP = make_tcp_connection(ip, port);
  if (*theDataSockP < 0) {
    return FTP_TCP_ERR;
  }

  // get server reply
  serverRetcode = ftp_get_retcode(theControlSock, tempBuf);
  if (serverRetcode == FTP_TCP_ERR) {
    close(*theDataSockP);
    return FTP_TCP_ERR;
  } 
  else if (serverRetcode != FTP_FILE_REQ_OK) {
    // unexpected FTP retcode for this command
    close(*theDataSockP);
    return FTP_FTP_ERR;
  } 
  else {
    return FTP_CMD_OK;
  }
}

int ftp_parse_list_out(int theDataSock,
		       unsigned long *theSizeP) {
  int err, status;
  char tempBuf[FTP_MAX_LINE];

  *theSizeP = 0;

  err = timeout_read_line(&status, theDataSock, tempBuf, FTP_MAX_LINE, FTP_TIMEOUT);
  if (err) {
    return FTP_TCP_ERR;
  }

  status = sscanf(tempBuf, "%*s %*s %*s %*s %ld", theSizeP);
  if (status != 1) {
    // try another format
    status = sscanf(tempBuf, "%*s %*s %*s %ld", theSizeP);
  }

  if (status != 1) {   
    // did not scan the size of the file
    return FTP_FTP_ERR;
  }
  else {
    return FTP_CMD_OK;
  }
}
