#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "utils.h"
#include "hash_check.h"
#include "leech_protocol.h"
#include "leech_utils.h"

static void hash_check_serve_client(hashCheck_t* hasher);

static char s_hashBuf[HASH_BUF_SIZE];

hashCheck_t* hash_check_create(context* c) {
  hashCheck_t* hasher;

  // If hashing is not enabled, don't do anything here
  if(!c->hash_enabled)
    return NULL;

  hasher = (hashCheck_t *)malloc(sizeof(hashCheck_t));

  if(hasher == NULL) {
    Dprintf("hash_check_create::malloc failed, not responding to hash requests\n");
    return NULL;
  }

  OpenSSL_add_all_digests();
  if(hasher == NULL) {
    perror("hash_check_create::malloc");
    exit(-1);
  }

  hasher->theContext = c;

  hasher->evp_md = EVP_get_digestbyname("md5");
  if(hasher->evp_md == NULL) {
    perror("hash_check_create::EVP_get_digestbyname");
    exit(-1);
  }
  hasher->hashQueue = Q_Create();
  if(hasher->hashQueue == NULL) {
    perror("hash_check_create::malloc");
    exit(-1);
  }

  hasher->servThread = (pthread_t *)malloc(sizeof(pthread_t));
  if(hasher->servThread == NULL) {
    perror("hash_check_create::malloc");
    exit(-1);
  }

  hasher->mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
  if(hasher->mutex == NULL) {
    perror("hash_check_create::malloc");
    exit(-1);
  }
  pthread_mutex_init(hasher->mutex, NULL);

  hasher->cond = (pthread_cond_t *)malloc(sizeof(pthread_cond_t));
  if(hasher->cond == NULL) {
    perror("hash_check_create::malloc");
    exit(-1);
  }
  pthread_cond_init(hasher->cond, NULL);

  if(pthread_create(hasher->servThread, NULL, (void *)hash_check_serve_client,
		    (void*)hasher) != 0) {
    perror("hash_check_create::pthread_create");
    exit(-1);
  }
  pthread_detach(*hasher->servThread);
  return hasher;
}

void hash_check_destroy(hashCheck_t* hasher) {
  pthread_cancel(*hasher->servThread);

  //free the other variables, bla bla
}

int hash_check_calc_hash(hashCheck_t* hasher, size_t offset, size_t len, 
			 unsigned char* digest, int* digest_len) {
  // md5sum had better point to an array of EVP_MAX_MD_SIZE bytes
  int rv, amt;
  int total = 0;
  EVP_MD_CTX ctx;

  EVP_DigestInit(&ctx, hasher->evp_md);
  while(total < len) {
    amt = (len - total > HASH_BUF_SIZE) ? HASH_BUF_SIZE : len - total;
    rv = leech_file_read(hasher->theContext, s_hashBuf, offset+total, amt);
    if(rv != amt) {
      Dprintf("hash_check_calc_hash: unable to read from file\n");
      return -1;
    }
    EVP_DigestUpdate(&ctx, s_hashBuf, amt);
    total += amt;
  }
  EVP_DigestFinal(&ctx, digest, digest_len);
  return 0;
}

void hash_check_handle_client(hashCheck_t* hasher, connection_t* con) {
  char tempBuf[30];
 
  if(hasher == NULL) {
    // We don't have hashing enabled so pretend we don't have this file
    leech_send_retcode(con,
		       LEECH_FILE_UNAVAIL,
		       LEECH_FILE_UNAVAIL_STR);
    connection_destroy(con);
    return;
  }

  Dprintf("HASH CHECKER: received connection from %s:%d\n",
	  connection_get_peer_addr(con, tempBuf, 30),
	  connection_get_peer_port(con));

  pthread_mutex_lock(hasher->mutex);
  Q_Enqueue(hasher->hashQueue, con);
  pthread_cond_signal(hasher->cond);
  pthread_mutex_unlock(hasher->mutex);
}

// This is the body of the hash checking thread
void hash_check_serve_client(hashCheck_t* hasher) {
  size_t offset, len;
  int digest_len = EVP_MAX_MD_SIZE;
  char urlBuf[LEECH_MAX_LINE];
  char digest[EVP_MAX_MD_SIZE];
  connection_t* con;
  int rv, err;
  
  printf("Starting hash check thread\n");
  while(1) {
    pthread_mutex_lock(hasher->mutex);
    while(Q_Empty(hasher->hashQueue))
      pthread_cond_wait(hasher->cond, hasher->mutex);
    con = Q_Dequeue(hasher->hashQueue);
    pthread_mutex_unlock(hasher->mutex);
    
    rv = leech_get_cmd(con, urlBuf, LEECH_MAX_LINE, &offset, &len);
    if(rv != LEECH_OK) {
      if(rv != LEECH_BYE) {
	Dprintf("hash_check_serve_client: unknown command: rv=%d\n", rv);
	connection_destroy(con);
      }
      else {
	Dprintf2(5, "hash_check_serve_client: graceful disconnect\n");
	connection_destroy(con);
      }
      continue;
    }
    
    if(strcmp(urlBuf, hasher->theContext->URL) != 0) {
      leech_send_retcode(con,
			 LEECH_FILE_UNAVAIL,
			 LEECH_FILE_UNAVAIL_STR);
      connection_destroy(con);
      continue;
    }
    
    // Make sure we have the requested bytes
    if(!blocksIsByteRangeSet(hasher->theContext->dlBitMap, offset, len)) {
      leech_send_retcode(con, LEECH_BYTES_UNAVAIL, LEECH_BYTES_UNAVAIL_STR);
      connection_destroy(con);
      continue;
    }
    
    // Calculate the hash for the requested portion of the file
    rv = hash_check_calc_hash(hasher, offset, len, digest, &digest_len);
    if(rv < 0) {
      leech_send_retcode(con,
			 LEECH_BYTES_UNAVAIL,
			 LEECH_BYTES_UNAVAIL_STR);
      connection_destroy(con);
      continue;
    }

    leech_send_retcode(con, LEECH_CMD_OK, LEECH_CMD_OK_STR);

    // Send the hash to the requester
    err = timeout_writeall(&rv, con->sock, digest, digest_len, LEECH_TIMEOUT);

    if(err || rv < digest_len) {
      Dprintf("HASH CHECKER: error in write: err=%d, status=%d\n", err, rv);
      connection_destroy(con);
      continue;
    }

    // All hash connections are persistent so requeue this connection
    pthread_mutex_lock(hasher->mutex);
    pthread_cond_broadcast(hasher->cond);
    Q_Enqueue(hasher->hashQueue, con);
    pthread_mutex_unlock(hasher->mutex);
    
  }
}
