#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/socket.h>
#include "hash_request.h"
#include "leech_protocol.h"
#include "utils.h"

hashRequest_t* hash_request_create(context* ctx) {
  hashRequest_t* ret = (hashRequest_t *)malloc(sizeof(hashRequest_t));
  if(ret == NULL) {
    perror("hash_request_create::malloc");
    exit(-1);
  }
  ret->numVerif = DEFAULT_HASH_VERIFICATIONS;

  // Set the context
  ret->ctx = ctx;

  // Get the md5 digest for hashing
  ret->evp_md = EVP_get_digestbyname("md5");
  if(ret->evp_md == NULL) {
    perror("hash_request_create::EVP_get_digestbyname");
    exit(-1);
  }

  ret->hashThread = (pthread_t *)malloc(sizeof(pthread_t));
  if(ret->hashThread == NULL) {
    perror("hash_request_create::malloc");
    exit(-1);
  }
  
  // Set all of the connections to NULL
  memset(ret->reqs, 0, sizeof(hashInfo_t*) * sizeof(ret->reqs));
  ret->numReqs = 0;
  
  ret->inputQueue = Q_Create();
  if(ret->inputQueue == NULL) {
    perror("hash_request_create::malloc");
    exit(-1);
  }
  
  // Set up our stuff for select()
  FD_ZERO(&ret->sockets);
  ret->maxFD = 0;

  ret->mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
  if(ret->mutex == NULL) {
    perror("hash_request_create::malloc");
    exit(-1);
  }
  pthread_mutex_init(ret->mutex, NULL);
  
  ret->cond = (pthread_cond_t *)malloc(sizeof(pthread_cond_t));
  if(ret->cond == NULL) {
    perror("hash_request_create::malloc");
    exit(-1);
  }
  pthread_cond_init(ret->cond, NULL);
  
  if(pthread_create(ret->hashThread, NULL, (void *)hash_request_run_thread, 
		    (void *)ret) < 0) {
    perror("hash_request_create::pthread_create");
    exit(-1);
  }
  pthread_detach(*ret->hashThread);
  return ret;
}

int hash_request_make_request(hashRequest_t* hasher, hashHead_t* req, int sock) {
  // A request should simply have req->offset and req->len filled in
  // The rest should be NULL / blank

  int num_reqs;
  int size, x;
  struct sockaddr_in addr;
  socklen_t addr_len = sizeof(addr);
  // The portion of the file referenced here is beyond the end of the file
  if(req->offset + req->len > hasher->ctx->fileSize)
    return -1;

  if(getpeername(sock,(struct sockaddr *)&addr, &addr_len) < 0) {
    perror("hash_request_make_request::getpeername");
    return -1;
  }
  pthread_mutex_lock(hasher->mutex);
  if(req->len > hasher->ctx->hash_maxRequestSize) {
    size = hasher->ctx->hash_maxRequestSize;
    // We should get at least 2 here
    num_reqs = ceil((double)req->len / hasher->ctx->hash_maxRequestSize);
    for(x=0; x<num_reqs-1; x++) {
      hashHead_t* new_req;

      new_req = (hashHead_t *)malloc(sizeof(hashHead_t));
      new_req->offset = req->offset + x*size;
      // Make sure we calc the correct len for the last piece
      new_req->len = size;
      new_req->src_addr = addr.sin_addr.s_addr; // Network byte order
      new_req->src_port = ntohs(addr.sin_port); // Host byte order
      Q_Enqueue(hasher->inputQueue, new_req);
    }
    // For the last piece
    req->offset += x*size;
    req->len -= x*size;
  }
  Q_Enqueue(hasher->inputQueue, req);
  // Wake up the processing thread
  pthread_cond_signal(hasher->cond);
  pthread_mutex_unlock(hasher->mutex);
  return 0;
}

int hash_request_send_requests(hashRequest_t* hasher, hashHead_t* req,
			       download** dst, int num) {
  int i, j;
  int rand, err;
  download* temp;
  connection_t* conn;
  char ipbuf[BUFLEN];
  
  // Randomly permute the elements of <dst>
  for(i=0; i<3; i++) {
    for(j=0; j<num; j++) {
      rand = abs(mrand48()) % num;
      // swap positions <x> and <rand>
      temp = dst[j];
      dst[j] = dst[rand];
      dst[rand] = temp;
    }
  }
  
  for(i=0; i<hasher->numVerif; i++) {
    if(hasher->numReqs < MAX_HASH_CONNECTIONS) {
      conn = connection_create(PROTO_HASH_CHECK, SLURPIE_VERSION);
      inet_ntop(AF_INET, &dst[i]->ip, ipbuf, BUFLEN);
      err = connection_connect(conn, ipbuf, dst[i]->port);
      if(err) {
	Dprintf("hash_request_send_requests:: %s:%d bad connection\n",
		ipbuf, dst[i]->port);
	connection_destroy(conn);
	continue;
      }
    }
  }
  return 0;
}

download * clients[DL_MAXNODESPERBLOCK];

void hash_request_run_thread(hashRequest_t* hasher) {

  hashHead_t* req;
  int found = DL_MAXNODESPERBLOCK;
  
  while(1) { 
    pthread_mutex_lock(hasher->mutex);
    while(Q_Empty(hasher->inputQueue))
      pthread_cond_wait(hasher->cond, hasher->mutex);
    req = Q_Dequeue(hasher->inputQueue);
    updateTreeSearch(hasher->ctx->updateTree, req->offset / hasher->ctx->blockSize, &found, clients);
    hash_request_send_requests(hasher, req, clients, found);
  }
}

