#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>

#include "utils.h"
#include "queue.h"
#include "protocol.h"
#include "leech_utils.h"
#include "leeches.h"
#include "leech_protocol.h"
#include "leech_server.h"

static void leech_server_serve_client(leech_server_t *theCtxtP);

static void leech_server_shutdown(leech_t *theArgP, int theExitStatus);


leech_server_t* leech_server_create(context *theUserParamsP) {
  int err, i;
  leech_server_t *retP = 
    (leech_server_t *) malloc(sizeof(leech_server_t));

  assert(theUserParamsP != NULL);
  assert(retP != NULL);

  memset(retP, 0, sizeof(leech_server_t));

  // pointer to the context aka user params
  retP->userParamsP = theUserParamsP;

  // create leeches queue
  retP->leechesQP = Q_Create();

  // create the cond variable
  retP->leechesQCondP = malloc(sizeof(pthread_cond_t));
  if (!(retP->leechesQCondP)) {
    perror("leech_server_create::malloc");
    abort();
  }
  pthread_cond_init(retP->leechesQCondP, NULL);

  // create the serving threads
  retP->threadsP = 
    malloc(sizeof(pthread_t *)*theUserParamsP->numOfLeechThreads);
  for (i = 0; i < theUserParamsP->numOfLeechThreads; ++i) {
    retP->threadsP[i] = malloc(sizeof(pthread_t));
    err = pthread_create(retP->threadsP[i],
			 NULL,
			 (void *) &leech_server_serve_client,
			 (void *) retP);
    if(err){
      perror("leech_server_create:: pthread_create");
      abort();
    }
    pthread_detach(*retP->threadsP[i]);
  }

  return retP;
}

void leech_server_destroy(leech_server_t *theCtxtP) {
  int i;

  // cancel threads
  for (i = 0; i < theCtxtP->userParamsP->numOfLeechThreads; ++i) {
    pthread_cancel(*theCtxtP->threadsP[i]);
    free(theCtxtP->threadsP[i]);
  }
  free(theCtxtP->threadsP);

  // TBD gracefule destroy leeches
  // no time right now ):
  // destroy the leeches queue
  Q_Destroy(theCtxtP->leechesQP);

  free(theCtxtP);

}


void leech_server_handle_client(leech_server_t *theCtxtP, 
				       connection_t *theConnectionP) {
  char tempBuf[30];
  leech_t *leechP;

  // trace printouts
  Dprintf("LEECH SERVER: received connection from %s:%d\n",
	  connection_get_peer_addr(theConnectionP, tempBuf, 30),
	  connection_get_peer_port(theConnectionP));

  // if we exceed the number of clients send service unavailable message
  if (Q_sizeTS(theCtxtP->leechesQP) > 
      theCtxtP->userParamsP->maxNumOfLeeches) {
    // number of clients exceeded, send service unavailable message
    leech_send_retcode(theConnectionP, 
		       LEECH_SERVICE_UNAVAIL, 
		       LEECH_SERVICE_UNAVAIL_STR);
    connection_destroy(theConnectionP);
  }
  else {
    // create leech "object" and start serving the client
    leechP = leechCreate(theConnectionP);
    // add leech "object" to the leeches list and signal the server threads
    pthread_mutex_lock(theCtxtP->leechesQP->lock);
    Q_Enqueue(theCtxtP->leechesQP, leechP); 
    pthread_cond_signal(theCtxtP->leechesQCondP);
    pthread_mutex_unlock(theCtxtP->leechesQP->lock);
  }
}

// this is body of leech serving threads
void leech_server_serve_client (leech_server_t *theCtxtP) {
  leech_t *leechP;
  // pointer to the buffer from where will serve
  char *fileBufP;
  // offset from which the file to be served
  size_t offset;
  // the number of bytes to be served
  size_t len;
  // URL name
  char urlBufP[LEECH_MAX_LINE]; 

  int rv,err;

  while (1) {
    // synchronize
    pthread_mutex_lock(theCtxtP->leechesQP->lock);
    while(Q_Empty(theCtxtP->leechesQP)) {
      pthread_cond_wait(theCtxtP->leechesQCondP, theCtxtP->leechesQP->lock);
    }
    leechP = Q_Dequeue(theCtxtP->leechesQP);
    pthread_mutex_unlock(theCtxtP->leechesQP->lock);

    // oppa got a leech
    // parse the command
    rv = leech_get_cmd(leechP->connectionP,
		       urlBufP,
		       LEECH_MAX_LINE,
		       &offset,
		       &len);
    if (rv != LEECH_OK) {
	    if(rv!=LEECH_BYE){
	      Dprintf("leech_server_serve_client: unknown command: rv=%d\n",rv);
	      leech_server_shutdown(leechP, -1);
	    } else {
		    Dprintf2(5,"leech_server_serve_client: graceful disconnect\n");
		    leech_server_shutdown(leechP,0);
	    }
      continue;
    }

    // check that we are actually downloading the same file
    if (strcmp(urlBufP, theCtxtP->userParamsP->URL) != 0) {
      leech_send_retcode(leechP->connectionP,
			 LEECH_FILE_UNAVAIL,
			 LEECH_FILE_UNAVAIL_STR);
      leech_server_shutdown(leechP, -1);
      continue;
    }

    // check that we have requested bytes
    if (!blocksIsByteRangeSet(theCtxtP->userParamsP->dlBitMap, offset, len)) {
      leech_send_retcode(leechP->connectionP,
			 LEECH_BYTES_UNAVAIL,
			 LEECH_BYTES_UNAVAIL_STR);
      leech_server_shutdown(leechP, -1);
      continue;
    }

    // allocate file buffer and read in the range
    fileBufP = (char *)malloc(len*sizeof(char));
    rv = leech_file_read(theCtxtP->userParamsP, fileBufP, offset, len);
    if (rv != len) {
      Dprintf("leech_server_serve_client: unable to read from file\n");
      leech_send_retcode(leechP->connectionP,
			 LEECH_BYTES_UNAVAIL,
			 LEECH_BYTES_UNAVAIL_STR);
      free(fileBufP);
      leech_server_shutdown(leechP, -1);
      continue;
    }

    // send command OK to the client
    leech_send_retcode(leechP->connectionP,
		       LEECH_CMD_OK,
		       LEECH_CMD_OK_STR);

    // send the bytes to the client
    Dprintf2(3,"LEECH SERVER write: %d-%d\n",offset,len);
    err=timeout_writeall(&rv,
		     leechP->connectionP->sock,
		     fileBufP,
		     len,
		     LEECH_TIMEOUT);
    free(fileBufP);
    if(err||rv<len){
	Dprintf("LEECH SERVER: error in write: err=%d,status=%d\n",err,rv);
	leech_server_shutdown(leechP, -1);
	continue;
    }
    if(leechP->connectionP->proto == PROTO_LEECH_PERSIST){ // requeue this connection
	    					 	// b/c it is persistent
	    Dprintf2(2,"Leech Server: requeuing persistent connection\n");
	    pthread_mutex_lock(theCtxtP->leechesQP->lock);
	    pthread_cond_broadcast(theCtxtP->leechesQCondP);
	    Q_Enqueue(theCtxtP->leechesQP,leechP);
	    pthread_mutex_unlock(theCtxtP->leechesQP->lock);
    } else
	    leech_server_shutdown(leechP, 0);    	// else close it
  }
}

void leech_server_shutdown(leech_t *theLeechP, int theExitStatus) {
  char tempBuf[30];

  Dprintf2(3,"leech_server_shutdown %d serving %s:%d\n", theExitStatus,
	  connection_get_peer_addr(theLeechP->connectionP, tempBuf, 30),
	  connection_get_peer_port(theLeechP->connectionP));
  // close connection
  connection_destroy(theLeechP->connectionP);
  // deallocate leech "object"
  leechDestroy(theLeechP);
}


void do_leech_server_tests() {
/*    context *userParamsP = (context *)malloc(sizeof(context)); */
/*    void *rvP; */

/*    userParamsP->destFileName = "blocks.txt"; */
/*    userParamsP->URL = "blocks.txt"; */
/*    userParamsP->maxNumOfLeeches = 3; */
/*    userParamsP->leeches = Q_Create(); */
/*    userParamsP->dlBitMap = blocksNew(1024, 4096*1024); */
/*    userParamsP->leechLocalPort = LEECH_DEFAULT_PORT; */

/*    // update some blocks */
/*    blocksSet(userParamsP->dlBitMap, 0); */
/*    blocksSet(userParamsP->dlBitMap, 11); */
/*    blocksSet(userParamsP->dlBitMap, 1024); */
/*    blocksSet(userParamsP->dlBitMap, 1025); */

/*    contextCreateLeechServer(userParamsP); */

/*    pthread_join(*(userParamsP->leechThreadP), &rvP); */
}
