#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/socket.h>
#include <errno.h>

#include "utils.h"
#include "queue.h"
#include "protocol.h"
#include "leech_utils.h"
#include "leeches.h"
#include "leech_protocol.h"
#include "leech_server_new.h"

// This is our unified buffer for reading blocks from the file
static char s_FileBufP[FILE_BUF_SIZE];

static void leech_server_new_serve_client(leech_server_t *theCtxtP);

static void leech_server_new_shutdown(leech_t *theArgP,
				      int theExitStatus);
static int leech_server_new_add_client(leech_server_t* theCtxtP,
				       leech_t* clientP);
static int leech_server_new_handle_read(leech_server_t* theCtxtP,
					leech_info_t* leechP, int index);
static int leech_server_new_get_file_blocks(leech_server_t* theCtxtP,
					    leech_info_t* leech_infoP,
					    char* buf);
static void leech_server_new_remove(leech_server_t* theCtxtP, int index, int code);
static void leech_server_new_reset_connection(leech_server_t* theCtxtP, int index);

void leech_server_new_remove(leech_server_t* theCtxtP, int index, int code) {
  int sock;
  int x, newmax = 0;

  sock = theCtxtP->leechesP[index]->leech->connectionP->sock;

  FD_CLR(sock, &theCtxtP->leechSet);

  if(sock == theCtxtP->maxFD) {
    // Our max fd left so we need to compute a new maximum
    for(x=0; x<theCtxtP->leechSize; x++) {
      int tmpsock;

      if(theCtxtP->leechesP[x] == NULL)
	continue;
      tmpsock = theCtxtP->leechesP[x]->leech->connectionP->sock;
      if(tmpsock > newmax && tmpsock != sock)
	newmax = tmpsock;
    }
    theCtxtP->maxFD = newmax;
  }
  leech_server_new_shutdown(theCtxtP->leechesP[index]->leech, code);
  if(index != theCtxtP->currLeeches - 1) {
    theCtxtP->leechesP[index] = theCtxtP->leechesP[theCtxtP->currLeeches-1];
    theCtxtP->leechesP[theCtxtP->currLeeches-1] = NULL;
  }
  else
    theCtxtP->leechesP[index] = NULL;
  theCtxtP->currLeeches--;
}

void leech_server_new_reset_connection(leech_server_t* theCtxtP, int index) {
  //printf("***************IN RESET\n");
  if(theCtxtP->leechesP[index]->leech->connectionP->proto == PROTO_LEECH_PERSIST) { 
    // requeue this connection b/c it is persistent
    Dprintf2(2,"Leech Server: requeuing persistent connection\n");
    // Reset the file position info for this connection
    theCtxtP->leechesP[index]->len = 0;
    theCtxtP->leechesP[index]->offset = 0;
    theCtxtP->leechesP[index]->currPos = 0;
  } else
    leech_server_new_remove(theCtxtP, index, 0);    	// else close it
}

leech_server_t* leech_server_new_create(context *theUserParamsP) {
  int err;
  leech_server_t *retP = 
    (leech_server_t *) malloc(sizeof(leech_server_t));

  assert(theUserParamsP != NULL);
  assert(retP != NULL);

  memset(retP, 0, sizeof(leech_server_t));

  // pointer to the context aka user params
  retP->userParamsP = theUserParamsP;

  // create leeches queue
  //retP->leechesQP = Q_Create();
  memset(&retP->leechSet, 0, sizeof(retP->leechSet));
  retP->leechSize = LEECH_ARR_SIZE;
  retP->leechesP = malloc(sizeof(leech_info_t *) * retP->leechSize);
  if(!retP->leechesP) {
    perror("leech_server_new_create::malloc");
    abort();
  }
  memset(retP->leechesP, 0, sizeof(leech_info_t *) * retP->leechSize);
  retP->maxFD = 0;
  retP->currLeeches = 0;
  retP->mutexP = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
  if(!retP->mutexP) {
    perror("leech_server_new_create::malloc");
    abort();
  }
  pthread_mutex_init(retP->mutexP, NULL);
  retP->threadsP = malloc(sizeof(pthread_t *));
  retP->threadsP[0] = malloc(sizeof(pthread_t));
  err = pthread_create(retP->threadsP[0], 
		       NULL, 
		       (void *)leech_server_new_serve_client, 
		       (void *)retP);
  if(err) {
    perror("leech_server_new_create::pthread_create");
    abort();
  }
  pthread_detach(*retP->threadsP[0]);
  return retP;
}

void leech_server_new_destroy(leech_server_t *theCtxtP) {
  int x;

  // cancel thread
  pthread_cancel(*theCtxtP->threadsP[0]);
  free(theCtxtP->threadsP[0]);
  free(theCtxtP->threadsP);
  for(x=0; x<theCtxtP->leechSize; x++) {
    if(theCtxtP->leechesP[x] != NULL) {
      connection_destroy(theCtxtP->leechesP[x]->leech->connectionP);
      leechDestroy(theCtxtP->leechesP[x]->leech);
      free(theCtxtP->leechesP[x]);
    }
  }
  free(theCtxtP->leechesP);
  free(theCtxtP);
}

int leech_server_new_add_client(leech_server_t* theCtxtP, leech_t* client) {
  int index = theCtxtP->currLeeches;
  int new_leech_size;

  if(theCtxtP->currLeeches == theCtxtP->leechSize) {
    // We need to resize our array to make more space
    leech_info_t** new_leeches;

    //theCtxtP->leechSize *= 2;
    //printf("***************Expanding array\n");
    Dprintf("leech_server_new_add_client:: realloc'ing\n");
    new_leeches = (leech_info_t **)realloc(theCtxtP->leechesP, theCtxtP->leechSize*2);
    if(!new_leeches) {
      // Couldn't resize our array
      perror("leech_server_new_add_client::realloc");
      return -1;
    }
    theCtxtP->leechesP = new_leeches;
    new_leech_size = theCtxtP->leechSize*sizeof(leech_info_t *);
    memset(theCtxtP->leechesP+new_leech_size, 0, new_leech_size);
    theCtxtP->leechSize *= 2;
  }
  //for(x=0; x<theCtxtP->leechSize; x++) {
  //if(theCtxtP->leechesP[x] == NULL) {
  theCtxtP->leechesP[index] = (leech_info_t *)malloc(sizeof(leech_info_t));
  if(!theCtxtP->leechesP[index]) {
    // Couldn't allocate space for new client
    perror("leech_server_new_add_client::malloc");
    return -1;
  }
  //memset(theCtxtP->leechesP[x], 0, sizeof(leech_info_t));
  
  theCtxtP->leechesP[index]->leech = client;
  theCtxtP->leechesP[index]->offset = 0;
  theCtxtP->leechesP[index]->len = 0;
  theCtxtP->leechesP[index]->currPos = 0;
  theCtxtP->currLeeches++;
  //printf("Adding new client to index %d\n", x);
  return 0;
}

int leech_server_new_handle_read(leech_server_t* theCtxtP,
				 leech_info_t* leech_infoP, int index) {
  int rv;
  size_t offset, len;
  char urlBufP[LEECH_MAX_LINE];

  rv = leech_get_cmd(leech_infoP->leech->connectionP,
		     urlBufP,
		     LEECH_MAX_LINE,
		     &offset,
		     &len);
  if (rv != LEECH_OK) {
    if(rv != LEECH_BYE){
      Dprintf("leech_server_serve_client: unknown command: rv=%d\n",rv);
      leech_server_new_remove(theCtxtP, index, -1);
    } else {
      Dprintf2(5,"leech_server_serve_client: graceful disconnect\n");
      leech_server_new_remove(theCtxtP, index, 0);
    }
    return -1;
  }
  
  // check that we are actually downloading the same file
  if (strcmp(urlBufP, theCtxtP->userParamsP->URL) != 0) {
      leech_send_retcode(leech_infoP->leech->connectionP,
			 LEECH_FILE_UNAVAIL,
			 LEECH_FILE_UNAVAIL_STR);
      leech_server_new_remove(theCtxtP, index, -1);
      return -1;
  }
  
  // check that we have requested bytes
  if (!blocksIsByteRangeSet(theCtxtP->userParamsP->dlBitMap, offset, len)) {
    leech_send_retcode(leech_infoP->leech->connectionP,
		       LEECH_BYTES_UNAVAIL,
		       LEECH_BYTES_UNAVAIL_STR);
    leech_server_new_remove(theCtxtP, index, -1);
    return -1;
  }
  
  // Keep track of the requested values
  leech_infoP->offset = offset;
  leech_infoP->len = len;
  leech_infoP->currPos = 0;

  // send command OK to the client
  leech_send_retcode(leech_infoP->leech->connectionP,
		     LEECH_CMD_OK,
		     LEECH_CMD_OK_STR);
  
  return 0;
}

int leech_server_new_get_file_blocks(leech_server_t* theCtxtP,
				     leech_info_t* leech_infoP, char* buf) {
  int rv;
  size_t fileLen = leech_infoP->len - leech_infoP->currPos;

  // If we still have more than our buffer space left to read, only read
  // as much as we can fit in our buffer
  
  if(fileLen > FILE_BUF_SIZE)
    fileLen = FILE_BUF_SIZE;

  rv = leech_file_read(theCtxtP->userParamsP, buf, 
		       leech_infoP->offset + leech_infoP->currPos,
		       fileLen);
  if (rv != fileLen) {
    Dprintf("leech_server_serve_client: unable to read from file\n");
    leech_send_retcode(leech_infoP->leech->connectionP,
		       LEECH_BYTES_UNAVAIL,
		       LEECH_BYTES_UNAVAIL_STR);
    //leech_server_new_shutdown(leech_infoP->leech, -1);
    return -1;
  }
  return fileLen;
}

void leech_server_new_handle_client(leech_server_t *theCtxtP, 
				    connection_t *theConnectionP) {
  char tempBuf[30];
  leech_t *leechP;
  int ret;
 
  // trace printouts
  Dprintf2(3,"NEW LEECH SERVER: received connection from %s:%d\n",
	  connection_get_peer_addr(theConnectionP, tempBuf, 30),
	  connection_get_peer_port(theConnectionP));

  // create leech "object" and start serving the client
  leechP = leechCreate(theConnectionP);
  // add leech "object" to the leeches list and signal the server threads
  pthread_mutex_lock(theCtxtP->mutexP);
  ret = leech_server_new_add_client(theCtxtP, leechP);
  pthread_mutex_unlock(theCtxtP->mutexP);
  if(ret < 0) {
    // We were unable to allocate memory to serve this client, so return
    // failure
    leech_send_retcode(theConnectionP, 
		       LEECH_SERVICE_UNAVAIL, 
		       LEECH_SERVICE_UNAVAIL_STR);
    connection_destroy(theConnectionP);
    return;
  }
  pthread_mutex_lock(theCtxtP->mutexP);
  FD_SET(theConnectionP->sock, &theCtxtP->leechSet);
  if(theConnectionP->sock > theCtxtP->maxFD) {
    theCtxtP->maxFD = theConnectionP->sock;
    //printf("The max fd is now %d\n", theCtxtP->maxFD);
  }
  pthread_mutex_unlock(theCtxtP->mutexP);
}

// this is body of the only leech serving thread
void leech_server_new_serve_client (leech_server_t *theCtxtP) {
  // pointer to the buffer from where will serve
  // offset from which the file to be served
  fd_set readSet, writeSet;
  int len, written, rv;
  struct timeval tv;

  while (1) {
    int x;
    // Copy our sets of connections so they don't get modified by select
    pthread_mutex_lock(theCtxtP->mutexP);
    memcpy(&readSet, &theCtxtP->leechSet, sizeof(fd_set));
    memcpy(&writeSet, &theCtxtP->leechSet, sizeof(fd_set));
    pthread_mutex_unlock(theCtxtP->mutexP);

    tv.tv_sec = 1;
    tv.tv_usec = 0;
    rv = select(theCtxtP->maxFD+1, &readSet, &writeSet, NULL, &tv);
    if(rv < 0) {
      perror("leech_server_new_serve_client::select");
      continue;
    }
    if(rv == 0)
      continue;

    pthread_mutex_lock(theCtxtP->mutexP);
    for(x=0; x<theCtxtP->leechSize; x++) {
      if(theCtxtP->leechesP[x] != NULL) {
	leech_t* leech = theCtxtP->leechesP[x]->leech;
	
	if(FD_ISSET(leech->connectionP->sock, &readSet)) {
	  if(theCtxtP->leechesP[x]->len == 0) {
	    //printf("A REad!\n");
	    if(leech_server_new_handle_read(theCtxtP, theCtxtP->leechesP[x], x) < 0) {
	      continue;

	    }
	  }
	}
	if(FD_ISSET(leech->connectionP->sock, &writeSet)) {
	  if(theCtxtP->leechesP[x]->len != 0) {
	    //printf("A write!\n");
	    // Make sure we have something to write
	    len = leech_server_new_get_file_blocks(theCtxtP, 
						   theCtxtP->leechesP[x], 
						   s_FileBufP);
	    if(len < 0)
	      leech_server_new_remove(theCtxtP, x, -1);
	    else {
	      written = send(leech->connectionP->sock, s_FileBufP, len, 0);
	      if(written < 0) {
		leech_server_new_remove(theCtxtP, x, -1);
		continue;
	      }
	      theCtxtP->leechesP[x]->currPos += written;
	      if(theCtxtP->leechesP[x]->currPos == theCtxtP->leechesP[x]->len) {
		// We are done sending this guy all the data he wanted
		leech_server_new_reset_connection(theCtxtP, x);
		Dprintf("leech_server_new: finished sending %d bytes\n",len);
	      }
	    }
	  }
	}
      }
    }
    pthread_mutex_unlock(theCtxtP->mutexP);
  }
}

void leech_server_new_shutdown(leech_t* leechP, int theExitStatus) {
  char tempBuf[30];

  Dprintf2(3,"leech_server_shutdown %d serving %s:%d\n", theExitStatus,
	  connection_get_peer_addr(leechP->connectionP, tempBuf, 30),
	  connection_get_peer_port(leechP->connectionP));
  // close connection
  connection_destroy(leechP->connectionP);
  // deallocate leech "object"
  leechDestroy(leechP);
}


void do_leech_server_new_tests() {
/*    context *userParamsP = (context *)malloc(sizeof(context)); */
/*    void *rvP; */

/*    userParamsP->destFileName = "blocks.txt"; */
/*    userParamsP->URL = "blocks.txt"; */
/*    userParamsP->maxNumOfLeeches = 3; */
/*    userParamsP->leeches = Q_Create(); */
/*    userParamsP->dlBitMap = blocksNew(1024, 4096*1024); */
/*    userParamsP->leechLocalPort = LEECH_DEFAULT_PORT; */

/*    // update some blocks */
/*    blocksSet(userParamsP->dlBitMap, 0); */
/*    blocksSet(userParamsP->dlBitMap, 11); */
/*    blocksSet(userParamsP->dlBitMap, 1024); */
/*    blocksSet(userParamsP->dlBitMap, 1025); */

/*    contextCreateLeechServer(userParamsP); */

/*    pthread_join(*(userParamsP->leechThreadP), &rvP); */
}
