#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <assert.h>
#include <math.h>

#include "utils.h"
#include "neighbor.h"
#include "neighborManager.h"
#include "updates.h"
#include "protocol.h"
#include "hash_check.h"

// Private Local functions and types
static void* managerThread(void* neighborManagerPtr);
static void* acceptThread(void* neighborManagerPtr);

extern pthread_mutex_t neighborMutex;

neighborManager* neighborManagerCreate(context *ctx) {
  neighborManager* rv = NULL;
  rv = (neighborManager*)malloc(sizeof(neighborManager));
  if (rv == NULL) {
    fprintf(stderr, "neighborCreate error: Unable to allocate a neighborManager structure\n");
  }
  rv->lock = (pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
  if (pthread_mutex_init(rv->lock, NULL) != 0) {
    fprintf(stderr, "neighborManagerCreate:pthread_mutex_init failed!\n");
    abort();
  }
  pthread_mutex_lock(rv->lock);

  rv->maxNumOfNeighbors = ctx->maxNumOfNeighbors;
  rv->neighbors = malloc(sizeof(neighbor*)*rv->maxNumOfNeighbors);
  rv->nNeighbors=0;
  assert(rv->neighbors);
  rv->refCount = 1;
  rv->state = 1;

  rv->daemon = connection_create(PROTO_NEIGHBOR, SLURPIE_VERSION);
  if (rv->daemon == NULL) {
    fprintf(stderr, "neighborsManagerCreate:connection_create failed\n");
    abort();
  }

  rv->condNumOfNeighbors = (pthread_cond_t*)malloc(sizeof(pthread_cond_t));
  if (pthread_cond_init(rv->condNumOfNeighbors, NULL) != 0) {
    fprintf(stderr, "neighborManagerCreate:pthread_cond_init failed!\n");
    abort();
  }
  rv->condNewNeighbor = (pthread_cond_t*)malloc(sizeof(pthread_cond_t));
  if (pthread_cond_init(rv->condNewNeighbor, NULL) != 0) {
    fprintf(stderr, "neighborManagerCreate:pthread_cond_init failed!\n");
    abort();
  }

  rv->ctx=ctx;	// hook back to the global context
  // Start the Manager thread. Detached
  if(pthread_create(&rv->managerThread, NULL, managerThread, rv) < 0) {
    printf("neighborManagerCreate:pthread_create: error creating manager thread\n");
    abort();
  }

  // Makes the neighbor server
  if (connection_listen(rv->daemon, 0) != 0) {
    printf("neighborManagerCreate:make_tcp_server: error opening listening socket\n");
    abort();
  }
  // Start the Accept Thread. Detached
  if (pthread_create(&rv->acceptThread, NULL, acceptThread, rv) < 0) {
    printf("neighborManagerCreate:pthread_create: error createing accept thread\n");
    abort();
  }

  pthread_mutex_unlock(rv->lock);

  return rv;
}

void neighborManagerDestroy(neighborManager* d) {
  pthread_mutex_lock(d->lock);
  if (d && --d->refCount == 0) {
    if (pthread_join(d->acceptThread, NULL)) {
      printf("neighborManagerDestroy:pthread_join error on acceptThread\n");
      abort();
    }
    if (pthread_join(d->managerThread, NULL)) {
      printf("neighborManagerDestroy:pthread_join error on managerThread\n");
      abort();
    }

    // Kill the daemon connection
    connection_destroy(d->daemon);

    // Now delete all the neighbors
    free(d->neighbors);	// assumes individual neighbors are freed elsewhere

    if(pthread_cond_destroy(d->condNumOfNeighbors)) {
      printf("neighborManagerDestroy:pthread_cond_destroy failed!\n");
      abort();
    }
    free(d->condNumOfNeighbors);

    if(pthread_cond_destroy(d->condNewNeighbor)) {
      printf("neighborManagerDestroy:pthread_cond_destroy failed!\n");
      abort();
    }
    free(d->condNewNeighbor);

    pthread_mutex_unlock(d->lock);
    if(pthread_mutex_destroy(d->lock)) {
      printf("neighborManagerDestroy:pthread_mutex_destroy failed!\n");
      abort();
    }
    free(d->lock);

    free(d);
  }
  else {
    d->state = 0;
    pthread_cond_broadcast(d->condNumOfNeighbors);
    pthread_cond_broadcast(d->condNewNeighbor);
    pthread_mutex_unlock(d->lock);
  }
}

int neighborManagerSetMaxNumOfNeighbors(neighborManager* n, unsigned short neo) {
  int rv = -1;
  neighbor ** neo_array;
  int i;
  pthread_mutex_lock(n->lock);
  if (n->state) {
    n->maxNumOfNeighbors = neo;	
    neo_array= malloc(sizeof(neighbor*)*n->maxNumOfNeighbors);	// new array
    assert(neo_array);
    for(i=0;i<n->nNeighbors;i++)	// copy over old to new
	    neo_array[i]=n->neighbors[i];
    n->neighbors=neo_array;
    rv = 0;
    // We don't have a thread to immediatly kill links.
    // What will happen is that the acceptThread will end up not accepting any new 
    // ones till enough bad links are dropped by the manager thread
    pthread_cond_signal(n->condNumOfNeighbors);
  }
  pthread_mutex_unlock(n->lock);
  return rv;
}
const char* neighborManagerGetAddr(neighborManager* n , char* addr, size_t len) {
  return connection_get_sock_addr(n->daemon, addr, len);
}
unsigned short neighborManagerGetPort(neighborManager* n) {
	if(n)
	  return connection_get_sock_port(n->daemon);
	else
	  return 0;		// this happens at initialization, before nm is initted
}
int neighborManagerAddPeer(neighborManager* n, connection_t* neo) {
  int rv = -1;
  pthread_mutex_lock(n->lock);
  if (n->state) { // If we're still running
    if (n->nNeighbors < n->maxNumOfNeighbors) {
      neighbor* t;
      if ((t = neighborCreate(neo)) != NULL) {
	rv = 0;
	n->neighbors[n->nNeighbors++]=t;
	pthread_cond_signal(n->condNewNeighbor);
      }
    }
  }
  pthread_mutex_unlock(n->lock);
  return rv;
}
int neighborManagerConnectToPeer(neighborManager*n , const char* host, unsigned short port) {
  int rv = 0;
  connection_t* neo = connection_create(PROTO_NEIGHBOR, SLURPIE_VERSION);
  rv = neo == NULL;
  rv = (rv) ?  rv : connection_connect(neo, host, port);	
  rv = (rv) ?  rv : neighborManagerAddPeer(n, neo); 
  return rv;
}

/* int neighborManagerRemovePeer(neighborManager *nm, neighbor *del) {
 * 	ASSUMES nm is LOCKED
 */

int neighborManagerRemovePeer(neighborManager *nm, neighbor *del) {
  int i;
  neighbor *n;
  /*
  while((n = (neighbor *) Dequeue(nm->neighbors)) && (n != first))
    if (n != del) {
      Enqueue(nm->neighbors, n);
      if (!first)
	first = n;
    }

  if (n && n == first)
    Enqueue(nm->neighbors, n);
    */
  for(i=0;i<nm->nNeighbors;i++){
	  if(nm->neighbors[i]==del)
		  break;
  }
  assert(i<nm->nNeighbors);		// make sure we found it
  n = nm->neighbors[i];

  pthread_mutex_lock(&neighborMutex);
  if(--n->ref == 0) {
    nm->nNeighbors--;
    if((i<nm->nNeighbors)&&(nm->nNeighbors>0))	// if this wasn't the last one
      nm->neighbors[i]=nm->neighbors[nm->nNeighbors];	// swap em
    connection_destroy(n->link);
    free(n);
  }
  pthread_mutex_unlock(&neighborMutex);
  return 0;
}

/* addRandomNeighbor(n);
 *       pick someone random from our update tree
 *       - assumes nm is NOT locked, as it is locked in neighborManagerAddPeer()
 */
int addRandomNeighbor(neighborManager *n){
	update * u;
	int err;
	char ipbuf[BUFLEN];
	context *c;
	connection_t *neo;
	c = n->ctx;
	pthread_mutex_lock(n->lock);
	u = getRandomGoodNeighborUpdate(c);
	if(u==NULL){
		pthread_mutex_unlock(n->lock);
		return 2;
	}
	neo =connection_create(PROTO_NEIGHBOR,SLURPIE_VERSION);
	inet_ntop(AF_INET,&u->ip,ipbuf,BUFLEN);
	//err = connection_connect_from_port(neo, ipbuf, u->port,neighborManagerGetPort(n));  // use funky connect calls
	err = connection_connect(neo, ipbuf, u->port);  // use funky connect calls
	if(err){
		Dprintf("addRandomNeighbor: tried to add %s:%d, failed with %d\n",
				ipbuf,u->port, err);
		updateTreeRevoke(n->ctx->updateTree,u->ip,u->port);
		pthread_mutex_unlock(n->lock);
		return 3;
	}
	free(u);
	pthread_mutex_unlock(n->lock);
	neighborManagerAddPeer(n,neo);	// slight race condition; could add same node twice :(
	return 0;

}



// Private Local functions
static void* managerThread(void* neighborManagerPtr) {
  neighbor **list;
  int count;
  // struct timeval now;
  // struct timespec timeout;
  context *c;

  neighborManager* n = (neighborManager*)neighborManagerPtr;
  c=n->ctx;
  pthread_mutex_lock(n->lock);
  list = malloc(sizeof(neighbor*)*n->maxNumOfNeighbors);
  if(!list){
    perror("managerThread::malloc");
    assert(list!=NULL);
  }
  n->refCount++;
  while(n->state) {
    // get new neighbors if we need them
    if(c->neighborChoice){
	    count= MAX(4,4 *log(n->nNeighbors));
	    Dprintf2(10," log(%d) == %d", n->nNeighbors,(int) log(n->nNeighbors));
	    if(count!=c->desiredNumOfNeighbors){
		    Dprintf("managerThread:: readjusting desired neighbors from %d to %d\n",
				    c->desiredNumOfNeighbors,count);
		    c->desiredNumOfNeighbors=count;
	    }
    }
    if((n->nNeighbors<c->desiredNumOfNeighbors) && n->state) {
      pthread_mutex_unlock(n->lock);
      Dprintf("trying to add a random neighbor : %d < %d\n", n->nNeighbors, c->desiredNumOfNeighbors);
      addRandomNeighbor(n);
      if(n->nNeighbors == 0)
	sleep(1);
      pthread_mutex_lock(n->lock);
      //gettimeofday(&now,NULL);
      //timeout.tv_sec = now.tv_sec + 1;
      //timeout.tv_nsec = now.tv_usec * 1000;
      //pthread_cond_timedwait(n->condNewNeighbor, n->lock,&timeout);
    }
    // get updates from people who have them
    if ((n->state)&&(n->nNeighbors>0)){
      pthread_mutex_unlock(n->lock);	// unlock in select
      count = neighborManagerSelect(n,list,n->maxNumOfNeighbors);
      pthread_mutex_lock(n->lock);
      if(count> 0)	// undid sodre's evil code; hopefully will run faster
		getAllUpdates(n->ctx,list,count);
    }
  }
  // Bid farewell to the neighbors on exit
  while(n->nNeighbors) {
    neighborManagerRemovePeer(n,n->neighbors[0]);
  }
  pthread_mutex_unlock(n->lock);
  neighborManagerDestroy(n);
  return NULL;
}

static void* acceptThread(void* neighborManagerPtr) {

  neighborManager* n = (neighborManager*)neighborManagerPtr;
  char buf[BUFLEN];
  pthread_mutex_lock(n->lock);
  n->refCount++;
  connection_get_sock_addr(n->daemon,buf,BUFLEN);
  Dprintf("Neighbor Manager Listenning on:  %s:%d\n",
	  buf,
	  connection_get_sock_port(n->daemon));
  while(n->state) {
      int new_state = 1;
      connection_t* neo;
      pthread_mutex_unlock(n->lock);
      neo = connection_accept(n->daemon);
      if (neo) {
	switch(neo->proto) {
	case PROTO_NEIGHBOR:
	  if ((n->nNeighbors >= n->maxNumOfNeighbors)
			  ||(neighborManagerAddPeer(n, neo) != 0) )
	    connection_destroy(neo);
	  else {
	    connection_get_peer_addr(neo,buf,BUFLEN);
	    Dprintf("New NeighborConnection: %s:%d\n",
		    buf,
		    connection_get_peer_port(neo));
	  }
	  break;
	case PROTO_LEECH:
	case PROTO_LEECH_PERSIST:
	  if(n->ctx->useMTLeechServ)
	    leech_server_handle_client(n->ctx->leechServerP, neo);
	  else
	    leech_server_new_handle_client(n->ctx->leechServerP, neo);
	  break;
	case PROTO_HASH_CHECK:
	  hash_check_handle_client(n->ctx->hashChecker, neo);
	  break;
	default:
	  connection_destroy(neo);
	}
      } 
      else 
	new_state = 0;
      pthread_mutex_lock(n->lock);
      if (new_state == 0) {
	printf("acceptThread: error accepting connection on neighborManager\n");
	n->state = 0;
	assert(0);
      }
    }
    // Wait till the size of the queue is less than the number of max neighbors

  pthread_mutex_unlock(n->lock);
  neighborManagerDestroy(n);		// there is ref count on this, so it's cool
  return NULL;
}



/* int neighborManagerSelect(neighborManager* n, neighbor **list, int max);
 * 	do select() on all of the connections
 */
int neighborManagerSelect(neighborManager* nm, neighbor **list, int listlen){
  int count=0;
  int i,fd,max;
  fd_set fds, except_fds;
  struct timeval tv;
  neighbor *n;

  pthread_mutex_lock(nm->lock);
  tv.tv_sec = NEIGHBORMANAGER_TIMEOUT;
  tv.tv_usec = 0;

  max = -1;
  FD_ZERO(&fds);
  FD_ZERO(&except_fds);
  // foreach neighbor, add them to the fd list
  /*
  while((n = (neighbor *) Dequeue(nm->neighbors))&&(n!=first)){
    if(!first) 
      first = n;
    Enqueue(nm->neighbors,n);*/
  for(i=0;i<nm->nNeighbors;i++){
    if(!nm->neighbors[i]->del_flag) {
      fd = neighborGetFD(nm->neighbors[i]);
      FD_SET(fd, &fds);
      FD_SET(fd, &except_fds);
      if(fd>max)
	max=fd;
      count++;
    }
  }
  max++;
  pthread_mutex_unlock(nm->lock);		// release lock before we go to sleep
#ifdef NOTIMEOUTS
  i = select(max, &fds, NULL, &except_fds, NULL);	
#else
  i = select(max, &fds, NULL, &except_fds, &tv);
#endif
  if(i<=0){
    return i;
  }
  count = 0;
  pthread_mutex_lock(nm->lock);		// get lock back before we edited queue
  // Don't count on select detecting the exception raising
  // file descriptors!
  for(i=0;i<nm->nNeighbors;i++){
    n=nm->neighbors[i];
    if(FD_ISSET(neighborGetFD(n), &fds))
      list[count++] = neighborRef(n);
    if (FD_ISSET(neighborGetFD(n), &except_fds))
      //neighborDestroy(n);
      neighborManagerRemovePeer(nm, n);
    if(count>=listlen)
      break;
  }

  pthread_mutex_unlock(nm->lock);
  return count;
}

/* int neighborManagerGetList(neighborManager *, neighbor ** list, int listlen);
 * 	export a list of the current neighbors
 * 	-- NOTE: this proc exists b/c the neighbor list wasn't always implemented
 * 		as a List!
 */


/*
int neighborManagerGetList(neighborManager *nm, neighbor ** list, int listlen){
  int count=0;
  assert(nm&&list);

  pthread_mutex_lock(nm->lock);
  for(count=0;count<nm->nNeighbors;count++){
    if(count>=listlen){
      pthread_mutex_unlock(nm->lock);
      return count;
    }
    list[count]=nm->neighbors[count];
  }
  pthread_mutex_unlock(nm->lock);
  return count;
}

*/


// Unit Test
void do_neighborManager_tests() {
  neighborManager* nm;
  context c;
  c.maxNumOfNeighbors=10;
  c.updateTree=updateTreeNew(100);
  //neighbor* n;
  nm = neighborManagerCreate(&c);

  assert(neighborManagerConnectToPeer(nm, "localhost", neighborManagerGetPort(nm)) == 0);	

  neighborManagerDestroy(nm);
}
