#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <assert.h>
#include <string.h>
#include <arpa/inet.h>

#include "source_get.h"
#include "utils.h"
#include "ftp_utils.h"

int parseURL(const char *theURLP,
		     char** theHostNameP,
		     unsigned short **thePortP,
		     char** theFileNameP) {
  int hostNameLen;
  int fileNameLen;

  char *ptr1, *ptr2;

  ptr1 = strchr(theURLP, '/');
  if (ptr1 == NULL) {
    Dprintf("%s:%d URL format... \n", __FILE__, __LINE__);
    return -1;
  }
  ++ptr1;
  if (*ptr1 != '/') {
    Dprintf("%s:%d URL format... \n", __FILE__, __LINE__);
    return -1;
  }
  ++ptr1;
  // ptr1 is pointing to the first char after ://X
  // ptr2 is pointing to ://hostname:port/
  ptr2 = strchr(ptr1, '/');
  if (ptr2 == NULL) {
    Dprintf("%s:%d URL format... \n", __FILE__, __LINE__);
    return -1;
  }

  // deal with hostname
  hostNameLen = ptr2-ptr1+1; // +1 for the '\0'
  *theHostNameP = (char *) malloc(hostNameLen);
  if ((*theHostNameP) == NULL) {
    Dprintf("%s:%d malloc fails... \n", __FILE__, __LINE__);
    return -1;
  }
  strncpy(*theHostNameP, ptr1, hostNameLen-1);
  (*theHostNameP)[hostNameLen-1] = '\0';
  // look for the port number
  ptr1 = strchr(*theHostNameP, ':');
  if (ptr1 == NULL) {
    // port information is not contained in the URL
    *thePortP = NULL;
  }
  else {
    *ptr1 = '\0';
    ++ptr1;
    *thePortP = (unsigned short *)malloc(sizeof(unsigned int));
    **thePortP = atoi(ptr1);
  }

  // deal with file name
  fileNameLen = strlen(ptr2) + 1;
  *theFileNameP = (char *)malloc(fileNameLen);
  if (*theFileNameP == NULL) {
    Dprintf("%s:%d malloc fails... \n", __FILE__, __LINE__);
    return -1;
  }
  strncpy(*theFileNameP, ptr2, fileNameLen-1);
  (*theFileNameP)[fileNameLen-1] = '\0';

  return 0;
}

static unsigned long ftpGetFileSize(char *theURLP) {
  char *hostNameP;
  unsigned short *portP;
  char *fileNameP;
  int controlSock, dataSock;
  unsigned long rv;

  parseURL(theURLP, &hostNameP, &portP, &fileNameP);
  if (portP == NULL) {
    // means no port info in URL, assign default FTP server port
    portP = (unsigned short *)malloc(sizeof(unsigned short));
    *portP = IPPORT_FTP;
  }

  controlSock = make_tcp_connection(hostNameP, *portP);
  if (controlSock < 0) {
    Dprintf("%s %d ftp problems...\n", __FILE__, __LINE__);
    return 0;
  }

  if (ftp_connect_cmd(controlSock) != FTP_CMD_OK) {
    close(controlSock);
    Dprintf("%s %d ftp problems...\n", __FILE__, __LINE__);
    return 0;
  }

  if (ftp_login_cmd(controlSock, "ftp", "XXX@") !=  FTP_CMD_OK) {
    close(controlSock);
    Dprintf("%s %d ftp problems...\n", __FILE__, __LINE__);
    return 0;
  }

  if (ftp_list_cmd(controlSock, fileNameP, &dataSock) !=  FTP_CMD_OK) {
    close(controlSock);
    Dprintf("%s %d ftp problems...\n", __FILE__, __LINE__);
    return 0;
  }

  ftp_parse_list_out(dataSock, &rv);

  close(dataSock);
  close(controlSock);
  free(hostNameP);
  free(portP);
  free(fileNameP);

  return rv;
}

static unsigned long wwwGetFileSize(char *theURLP) {
  char *hostNameP;
  unsigned short *portP;
  char *fileNameP;
  char ip[BUFLEN];
  char tmpBuf[BUFLEN];
  int dataSock;
  unsigned long rv = 0;
  int err, count, status;

  parseURL(theURLP, &hostNameP, &portP, &fileNameP);
  if (portP == NULL) {
    // means no port info in URL, assign default FTP server port
    portP = (unsigned short *)malloc(sizeof(unsigned short));
    *portP = 80;
  }

  dataSock = make_tcp_connection(hostNameP, *portP);	// get connection
  hostnameToIpString(hostNameP, ip, BUFLEN);
  if (dataSock < 0) {
    Dprintf("%s %d Connection Failure host '%s' port '%u'\n", 
		__FILE__, __LINE__,hostNameP,*portP);
    return 0;
  }

  count=snprintf(tmpBuf, BUFLEN, "GET %s HTTP/1.1%s"
		 "Host: %s%s"
		 "Range: bytes=%d-%d%s%s",
		 fileNameP, CRLF, ip, CRLF, 0,
		 0, CRLF, CRLF);

  //Dprintf("SENDING:%s%s----\n",CRLF,tmpBuf);

  err=timeout_writeall(&status,dataSock,tmpBuf,count,GETTIMEOUT);

  if(err) {
    close(dataSock);
    Dprintf("%s %d http problems...\n", __FILE__, __LINE__);
    return 0;
  }

  if(status<count) {
    close(dataSock);
    Dprintf("%s %d http problems...\n", __FILE__, __LINE__);
    return 0;	// short write; must have died
  }

  while(!err){
    err = timeout_read_line(&status, dataSock, tmpBuf, BUFLEN, GETTIMEOUT);
    if (err || (status <= 0)) 
      return 0;
    // Dprintf(">%s\n",tmpBuf);	
    if(!strncmp(CONTENT_RANGE, tmpBuf, strlen(CONTENT_RANGE)))
      sscanf(tmpBuf, CONTENT_RANGE": bytes %*d-%*d/%ld", &rv);
    if(isBlankLine(tmpBuf))
      break;
  }

  close(dataSock);
  free(hostNameP);
  free(portP);
  free(fileNameP);

  return rv;
}

// will query the source for the file size
unsigned long getFileSize(char *theURLP) {
  assert(theURLP);
  if(!strncasecmp(theURLP, FTP_PROTO, strlen(FTP_PROTO)))
    return ftpGetFileSize(theURLP);
  if(!strncasecmp(theURLP, WWW_PROTO, strlen(WWW_PROTO)))
    return wwwGetFileSize(theURLP);
  return -1;	// protocol not found
}

/* source_get():
 * 	demux between the protocol specific gets
 */
int source_get(context *c, char * buf, int buflen, 
		unsigned int offset, unsigned int len){
	assert(c);
	// Dprintf("source_get: bytes: %u-%u\n", offset, offset + len - 1);
	if(!strncasecmp(c->URL,FTP_PROTO, strlen(FTP_PROTO)))
		return ftp_get(c,buf,buflen, offset, len);
	if(!strncasecmp(c->URL,WWW_PROTO, strlen(WWW_PROTO)))
		return www_get(c,buf,buflen, offset, len);
	return -1;	// protocol not found
}

/* ftp_get():
 * 	get file c->URL , but only bytes between offset and offset+len
 * 	return 0 on success
 */	
int ftp_get(context *c, 
	    char * buf, 
	    int buflen, 
	    unsigned int offset,
	    unsigned int len)
{

  char * hostname;
  int hostnamelen;
  char * filename;
  int filenamelen;
  char *ptr,*ptr2;
  int err; 


  ptr = strchr(c->URL,'/');	// first slash
  if(ptr==NULL)
    return 1;		// no slash
  ptr++;
  if(*ptr != '/')
    return 1;		// missing second slash
  ptr++;
  ptr2=strchr(ptr,'/');
  hostnamelen=ptr2-ptr+1;		// + one for the NULL
  hostname = malloc(hostnamelen);
  filenamelen=strlen(ptr2)+1;		
  filename = malloc(filenamelen);
  if(!hostname|| !filename){
    perror("ftp_get: malloc");
    return 2;
  }
  strncpy(hostname,ptr,hostnamelen);  // might include ":port"
  hostname[hostnamelen-1] = '\0';
  strncpy(filename,ptr2,filenamelen);  
  filename[filenamelen-1] = '\0';
  // TBD Elena is :port applicable to FTP?
  //if((ptr=strchr(hostname,':'))){
  //  *ptr=0;
  //  ptr++;
  //  port=atoi(ptr);
  //}

  //Dprintf("ftp_get host='%s' file='%s'\n",
//	  hostname,
//	  filename);

  err=do_ftp_get(hostname,filename,buf,buflen,offset,len);

  free(hostname);
  free(filename);

  return err;	
}

int do_ftp_get(char * hostname, 
	       char * filename, 
	       char * buf, 
	       int buflen, 
	       unsigned int offset, 
	       unsigned int len)
{
  int controlSock;
  int dataSock;
  int err, status;

  // FTP control connection
  controlSock = make_tcp_connection(hostname, IPPORT_FTP);
  if (controlSock < 0) {
    return controlSock;
  }

  if (ftp_connect_cmd(controlSock) != FTP_CMD_OK) {
    close(controlSock);
    return -1;
  }

  if (ftp_login_cmd(controlSock, "ftp", "XXX@") !=  FTP_CMD_OK) {
    close(controlSock);
    return -1;
  }

  if (ftp_retr_cmd(controlSock, 'I', filename, offset, &dataSock) !=  FTP_CMD_OK) {
    close(controlSock);
    return -1;
  }

  err = timeout_readall(&status, dataSock, buf, len, FTP_TIMEOUT);

  close(dataSock);
  close(controlSock);

  if ((status < len) || err) {
    return -1;
  } 
  else {
    return 0;
  }
}

/* www_get():
 * 	get file c->URL , but only bytes between offset and offset+len
 * 	return 0 on success
 *	anything else is an error
 *
 * 	ignore user/password stuff in URL for now... unless it just works ;)
 */
	
int www_get(context *c, char * buf, int buflen, unsigned int offset,
	unsigned int len){
	char * hostname;
	int hostnamelen;
	unsigned int port=80;
	char * filename;
	int filenamelen;
	char *ptr,*ptr2;
	int err;

	ptr = strchr(c->URL,'/');	// first slash
	if(ptr==NULL)
		return 1;		// no slash
	ptr++;
	if(*ptr != '/')
		return 1;		// missing second slash
	ptr++;
	ptr2=strchr(ptr,'/');
	hostnamelen=ptr2-ptr+1;		// + one for the NULL
	hostname = malloc(hostnamelen);
	filenamelen=strlen(ptr2)+1;		
	filename = malloc(filenamelen);
	if(!hostname|| !filename){
		perror("www_get: malloc");
		return 2;
	}
	strncpy(hostname,ptr,hostnamelen);  // might include ":port"
	strncpy(filename,ptr2,filenamelen);  
	filename[filenamelen-1]=0;	// superstition :)
	hostname[hostnamelen-1]=0;

	if((ptr = strchr(hostname, ':'))){
		*ptr = 0;
		ptr++;
		port = atoi(ptr);
	}

	//Dprintf("www_get host='%s' file='%s' port='%d'\n",
	//		hostname,
	//		filename,
	//		port);
	err=do_www_get(hostname,filename,port,buf,buflen,offset,len);
	free(hostname);
	free(filename);
	return err;	
}
	
/* int do_www_get(char * hostname, char * filename, unsigned short port,
 *                 char * buf, int buflen, unsigned int offset, unsigned int len);
 *       the thinger that actually does the connection
 */

int do_www_get(char * hostname, char * filename, unsigned short port,
		char * buf, int buflen, unsigned int offset, unsigned int len){
	int sock;
	char tmpbuf[BUFLEN];
	char ip[BUFLEN];
	int count,err,status;
	unsigned int contentLength=0;

	sock = make_tcp_connection(hostname,port);	// get connection
	hostnameToIpString(hostname,ip,BUFLEN);
	if(sock<0)
		return sock;
	count=snprintf(tmpbuf,BUFLEN,"GET %s HTTP/1.1%s"
			"Host: %s%s"
			"Range: bytes=%d-%d%s%s",
			filename, CRLF, ip, CRLF, offset,
			offset+len-1,CRLF,CRLF);

	//Dprintf("SENDING:%s%s----\n",CRLF,tmpbuf);
	err=timeout_writeall(&status,sock,tmpbuf,count,GETTIMEOUT);
	if(err) 
		return err;
	if(status<count)
		return status-count;	// short write; must have died

	while(!err){
		err = timeout_read_line(&status,sock,tmpbuf,BUFLEN,GETTIMEOUT);
		if(err) 
			return err;
		if(status<0)
			return status;
		if(status==0)
			return -1;
		//Dprintf(">%s\n",tmpbuf);	
		if(!strncmp(CONTENT_LENGTH,tmpbuf,strlen(CONTENT_LENGTH)))
			sscanf(tmpbuf,CONTENT_LENGTH": %d",&contentLength);
		if(isBlankLine(tmpbuf))
			break;
	}
	if(err) 
		return err;

	err= timeout_readall(&status,sock,buf,contentLength,GETTIMEOUT);
	close(sock);
	return err;
}

/*         int hostnameToIpString(hostname,ip,BUFLEN);
 *	turns "www.cs.umd.edu" -> "128.8.128.160"
 */

int hostnameToIpString(char * hostname,char * ip,int len){

	struct hostent h, *hptr;
	int err;
	char tmpbuf[HBUFLEN];
	struct in_addr i;

	gethostbyname_r(hostname, &h, tmpbuf, HBUFLEN, &hptr, &err);
	if(!hptr){
		*ip=0;
		return -1;
	}
	memcpy(&i.s_addr,hptr->h_addr, sizeof(i.s_addr));
	return inet_ntop(AF_INET,&i,ip,len)==NULL;
}

/*	isBlankLine(str)
 *		return 1 if str is composed of only whitespace 
 *		0 otherwise
 */
int isBlankLine(char * str){
	char *ptr=str;

	while(*ptr){
		switch(*ptr){
			case ' ':
			case '\r':
			case '\n':
			case '\t': 
				break;
			default:
				   return 0;
		}
		ptr++;
	}
	return 1;
}

int do_source_get_tests(){
  char buf[32*1024];	// 32k
  context *c;
  int err;

  c = malloc(sizeof(context));
  c->URL="http://www.cs.umd.edu/~mount/pubs.html";
  err=source_get(c,buf,32*1024,0,32);
  printf("do_source_get_tests:: Got: %s\n",buf);
  //assert(!err);
  if (err) {
    perror("do_source_get_tests:: http get failed...\n");
  }

  Dprintf("file size is : %ld\n", getFileSize(c->URL));

  //c->URL="http://www.myleft.net/index.html";
  //err=source_get(c,buf,32*1024,0,32*1024-1);
  //assert(!err);


  c->URL = "ftp://ftp.gnu.org/gnu/zlibc/zlibc-0.9j.tar.gz";
  err = source_get(c, buf, 32*1024,5,10);
  //assert(!err);
  if (err) {
    perror("do_source_get_tests:: ftp get failed...\n");
  }
  else {
    printf("ftp get:: Got: %s\n", buf);
  }


  Dprintf("file size is : %ld\n", getFileSize(c->URL));

  //free(c);
  return 0;
}
