#include<stdio.h>
#include<stdlib.h>
#include <assert.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <errno.h>



#include "updateTree.h"
#include "utils.h"


// local functions
void updateTreeProcessUpdate(UpdateTree *);
UpdateNode * UpdateNodeLookup(UpdateTree * ut,unsigned int ip,unsigned short port);
UpdateNode * updateNodeCreate(unsigned int ip, unsigned short port, blocks * b);
int updateTreeDoReplacement(UpdateTree * ut,unsigned int ip,unsigned short port,blocks *b);

/* updateTreeNew():
 * 	initialize an updateTree and return it
 */


UpdateTree * updateTreeNew(int maxLeaves){
	UpdateTree *ret;
	int i;

	assert(maxLeaves>0);
	ret = malloc(sizeof(UpdateTree));
	if(!ret){
		perror("updateTreeInit:: malloc");
		return NULL;
	}
	ret->maxLeaves = maxLeaves;
	ret->numLeaves = 0;
	ret->root = NULL;
	ret->leaves = malloc(sizeof(UpdateNode* )*maxLeaves);
	if(!ret->leaves){
		free(ret);
		perror("updateTreeInit:: malloc");
		return NULL;
	}

	ret->lock = malloc(sizeof(pthread_mutex_t));
	if(!ret->lock){
		free(ret->leaves);
		free(ret);
		perror("updateTreeInit:: malloc");
		return NULL;
	}
	for(i=0;i<maxLeaves;i++)
		ret->leaves[i]=NULL;

	pthread_mutex_init(ret->lock,NULL);	 // initialize the lock


	return ret;
}

/* int updateTreeFree(UpdateTree *);
 * 	free all resources in the updateTree
 */
int updateTreeFree(UpdateTree * ut){
	assert(ut);
	updateTreeTearDownTree(ut);
	free(ut->leaves);
	free(ut->lock);
	free(ut);
	return 0;
}



/* int updateTreeUpdate(UpdateTree *, unsigned int ip, unsigned
 *                 short port, blocks *b);
 *	add an update to the updateTree
 *		- if the entry exists, update it
 *		- else:
 *			if the tree isn't full, add it
 *			else: apply replacement rule
 */

int updateTreeUpdate(context * ctx, UpdateTree *ut, unsigned int ip, unsigned
		                short port, blocks *b){
	UpdateNode *n;
	int diff;

	assert(ut);
	pthread_mutex_lock(ut->lock);
	n = UpdateNodeLookup(ut,ip,port);
	if(n){					// case 1
		// TODO: are we really supposed to free B?
		blocks* mergedblocks;
		diff = blocksCmp(b,n->data);
		if(diff<=0){	// nothing new
			Dprintf2(2,"Ignoring old block from %d:%u: diff=%d\n",ip,port,diff);
			blocksFree(b);
			pthread_mutex_unlock(ut->lock);	
			return 0;
		}
		if (n->revoked)
			mergedblocks = blocksMerge(n->data, n->data);
		else 
			mergedblocks = blocksMerge(b, n->data);
		blocksFree(b);
		blocksFree(n->data);
		n->data = mergedblocks;
		updateTreeProcessUpdate(ut);	// propagate the update
		pthread_mutex_unlock(ut->lock);
		if((ctx->downloadType==DOWNLOADTYPE_ADAPTIVE)&&
				(ctx->ac)){	// tell the scheduler
			Dprintf2(5,"updateTree signalling active scheduler\n");
			pthread_mutex_lock(ctx->ac->lock);	// something new arrived
			pthread_cond_signal(ctx->ac->cond);
			pthread_mutex_unlock(ctx->ac->lock);
			Dprintf2(5,"updateTree 		DONE signalling active scheduler\n");
		}
		return 0;
	}
	if(ut->numLeaves<ut->maxLeaves){
		ut->leaves[ut->numLeaves] = updateNodeCreate(ip, port, b);
		assert(ut->leaves[ut->numLeaves]);
		ut->numLeaves++;
		updateTreeProcessUpdate(ut);	// propagate the update
		pthread_mutex_unlock(ut->lock);
		return 0;
	}
	updateTreeDoReplacement(ut,ip,port,b);
	updateTreeProcessUpdate(ut);	// propagate the update
	pthread_mutex_unlock(ut->lock);
	return 0;
}

/* UpdateNodeLookup(ut,ip,port):
 * 	return a ptr to the UpdateNode that matches (ip,port)
 * 		NULL if it doesn't exist
 * 	-- linear search should be fine for now.. .update later if it's a problem
 * 	-- assumes tree is locked
 */

UpdateNode * UpdateNodeLookup(UpdateTree * ut,unsigned int ip,unsigned short port){
	int i;
	assert(ut);
	for(i=0;i<ut->numLeaves;i++){
		if((ut->leaves[i]->ip == ip)&&(ut->leaves[i]->port == port)){
			return ut->leaves[i];
		}
	}
	return NULL;
}

/* updateNodeCreate(ip, port, b):
 * 	allocate and fill in an UpdateNode struct
 */

UpdateNode * updateNodeCreate(unsigned int ip, unsigned short port, blocks * b){
	UpdateNode * ret;

	ret = malloc(sizeof(UpdateNode));
	if(!ret){
		perror("UpdateNodeCreate::malloc");
		return NULL;
	}
	ret->busy = 0;
	ret->revoked = 0;
	ret->ip = ip;
	ret->port = port;
	ret->data = b;
	blocksIncRef(b);
	ret->left = ret->right = NULL;
	ret->next =  NULL;
	//ret->parent= NULL;

	return ret;
}

/* int updateNodeFree(UpdateNode *):
 * 	free's the resources in an updateNode
 */

int updateNodeFree(UpdateNode * un){
	assert(un);
	assert(un->data);
	blocksFree(un->data);
	free(un);
	return 0;
}

/*  int updateTreeSearch(UpdateTree *, int block, int * count, download ** dls);
 *  	- start at the root, and return a set of hosts that have the given block
 */

int updateTreeSearch(UpdateTree *ut, int block, int * count, download ** dls){
	int found;
	pthread_mutex_lock(ut->lock);
	found=updateTreeSearch2(ut->root, block, *count, 0, dls);
	pthread_mutex_unlock(ut->lock);
	*count=found;
	return found;
}

/* updateTreeSearch2() recursive part of updateTreeSearch()
 * 	returns the number found
 */

int updateTreeSearch2(UpdateNode *n, int block, int max, int curr, download ** dls){
	int found=0;
	if(n == NULL)
		return 0;
	if(curr>=max)			// buffer full
		return 0;	

	if(!blocksIsSet(n->data,block))	// that block is not know about
		return 0;
	if((n->left==NULL) && (n->right==NULL) && !n->busy){	// if we are a leaf and not busy
		dls[curr] = downloadCreateByIP(n->ip,n->port);
		return 1;
	}	

	// recurse
	found += updateTreeSearch2(n->left,block, max, curr, dls);
	curr+= found;
	found += updateTreeSearch2(n->right,block, max, curr, dls);
	return found;
}
		
	
/* updateTreeDoReplacement():
 * 	we have the maximum number of allowed state, so this is our cache replacement
 * 	policy	: for right now, just replace uniformly at random
 */

// STUB
int updateTreeDoReplacement(UpdateTree * ut,unsigned int ip,unsigned short port,blocks *b){
	ut=ut;
	ip=ip;
	port=port;
	b=b;
	return 0;
}

/* void updateTreeProcessUpdate(UpdateTree *);
 * 	rebuild the block location structure
 */

// STUB

void updateTreeProcessUpdate(UpdateTree *ut){
	int i,count;
	UpdateNode * un,*prev, *parent,*first;
	blocks *b;

	if(ut->numLeaves<1){
		assert(ut->root ==NULL);
		return;
	}
	updateTreeTearDownTree(ut);	// leaves us with ut->root=NULL

	count=ut->numLeaves;
	un = ut->leaves[0];
	while(count>1){		// per row
		first=NULL;
		prev=NULL;
		for(i=0;i<count;i+=2){	// every other node in a row
			assert(un);
			if(un->next){	// merge two blocks
				b=blocksMerge(un->data,un->next->data);
				parent = updateNodeCreate(0,0, b);
				parent->left = un;
				parent->right = un->next;
				if(prev){
					prev->next=parent;
				}
				// IMPORTANT!!!! iterate to the next set of two
				un = un->next->next;
				//un->parent=un->next->parent=parent;
			} else {	// just promote the existing node on a safe way
				b = blocksMerge(un->data, un->data);
				parent=updateNodeCreate(0,0, b);
				parent->left = un;
				parent->right = NULL;
				if(prev){
					prev->next=parent;
				}
				un = un->next; // should be null... and next iteration should not occurr
			}
			if(!first)	// save the first node in the new row
				first=parent;
			prev = parent;
		}
		if(count%2)
			count=1+count/2;
		else
			count=count/2;
		un=first;
				
	}
	ut->root = un;
}


/* updateTreeTearDownTree(ut);
 * 	free() the existing tree structure, and then clean it up
 * 	for the updateTreeProcessUpdate function
 */

int updateTreeTearDownTree(UpdateTree* ut){
	int i;
	assert(ut);
	updateTreeTearDownTree2(ut->root);

	for(i=0;i<ut->numLeaves-1;i++){
		ut->leaves[i]->next = (struct UpdateNode *) ut->leaves[i+1];
	}
	if(ut->numLeaves!= 0)
		ut->leaves[ut->numLeaves-1]->next=NULL;

	ut->root=NULL;
	return 0;
}

/* int updateTreeTearDownTree2(UpdateNode * );
 * 	recursively free any node that is not a child
 */
int updateTreeTearDownTree2(UpdateNode * un){

	if(un==NULL) 
		return 0;		// empty
	if((un->left==NULL)&&(un->right==NULL))	
		return 0;		// don't free the leaves
	updateTreeTearDownTree2(un->left);
	updateTreeTearDownTree2(un->right);
	updateNodeFree(un);
	return 0;
}

/* neighbor * getRandomGoodNeighborUpdate(struct context *c);
 * 	return a random neighbor (in an struct update)  
 * 	that is not already in our neighbor list
 * 	and that is not us
 */
update * getRandomGoodNeighborUpdate(struct context *c){
	update *ret=NULL;
	UpdateNode * u=NULL;
	UpdateTree * ut= c->updateTree;
	neighborManager * nm = c->neighborMgr;
	unsigned short localPort;
	int start,i,j,k;
	
	localPort = neighborManagerGetPort(nm);

	if(ut->numLeaves<1)	// we are the only ones in our tree, so no good neighbors
		return NULL;	// 	exist
	start = abs(mrand48())%ut->numLeaves;	// pick random divider
	//iterations=0;
	for(i=0; i<ut->numLeaves; i++) {
	  j = (i+start) % ut->numLeaves;
	  if(ut->leaves[j]->revoked)
	    continue;
	  if((ut->leaves[j]->ip==c->localIp)	// skip ourselves
	     &&(ut->leaves[j]->port==localPort))
	    continue;
	  for(k=0;k<nm->nNeighbors;k++){
	    if((nm->neighbors[k]->ip==ut->leaves[j]->ip)&&
	       (nm->neighbors[k]->remote_nm_port==ut->leaves[j]->port))
	      break;
	  }
	  if(k<nm->nNeighbors)
	    continue;	// this is one of our existing neighbors
	  u=ut->leaves[j];
	  ret = malloc(sizeof(update));
	  assert(ret);
	  ret->ip=u->ip;
	  ret->port=u->port;
	  return ret;
	}
	return NULL;
	/*while(iterations<2){
		if(iterations==0){	// after divider
			i=start;
			stop=ut->numLeaves;
		} else {		// before divider
			i=0;
			stop=start;
		}
		for(;i<stop;i++){
			if(ut->leaves[i]->revoked)	// skip revoked ppl
				continue;
			if((ut->leaves[i]->ip==c->localIp)	// skip ourselves
					&&(ut->leaves[i]->port==localPort))
				continue;
			for(j=0;j<nm->nNeighbors;j++){
				if((nm->neighbors[j]->ip==ut->leaves[i]->ip)&&
						(nm->neighbors[j]->remote_nm_port==ut->leaves[i]->port))
					break;
			}
			if(j<nm->nNeighbors)
				continue;	// this is one of our existing neighbors
			u=ut->leaves[i];
			break;		// if we got this far, then this is a good neighbor
		}
		if(u!=NULL)
			break;
		iterations++;
	}

	if(u==NULL)	// did we find anyone?
		return NULL;
	ret = malloc(sizeof(update));
	assert(ret);
	ret->ip=u->ip;
	ret->port=u->port;
	return ret;*/
}


/* update * getRandomUpdate(UpdateTree *);
 * 	return a list of filled out update structs choosen uniformly
 * 	at random from the updateTree
 */

int  getRandomUpdates(UpdateTree *ut,context *c,update ** uList, int max){
	int start,i,j;
	int count,nLeaves;

	nLeaves = updateTreeCountNodes(c);
	if(nLeaves<=0)	// short cut if we have nothing
		return 0;

	start = abs(mrand48())%ut->numLeaves;
	/*
	i=start;
	do{
		if(ut->leaves[i]->revoked){	// don't return a revoked block
			i++;
			if(i>=ut->numLeaves)
				i=0;
		}
		ret->ip = ut->leaves[i]->ip;
		ret->port = ut->leaves[i]->port;
		ret->data = ut->leaves[i]->data;
		blocksIncRef(ut->leaves[i]->data);	// increment reference count on the blocks
		ret->type = c->blockRule;
		return ret;
	} while(i!=start);
	free(ret);
	return NULL;		// didn't find anyone
	*/
	count=0;
	for(i=0;i<ut->numLeaves;i++){
		j=(i+start)%ut->numLeaves;
		if(ut->leaves[j]->revoked)
			continue;
		uList[count]=malloc(sizeof(update));
		uList[count]->ip=ut->leaves[j]->ip;
		uList[count]->port=ut->leaves[j]->port;
		uList[count]->data=ut->leaves[j]->data;
		uList[count]->type = c->blockRule;
		blocksIncRef(ut->leaves[j]->data);
		count++;
		if(count>=max)
			return count;
	}
	return count;
}

/* int updateTreeFindMissingBlock(UpdateTree *);
 * 	return an index for a block that is not in the updateTree,
 * 	between the range [start:stop] (inclusive!)
 * 	i.e. one that we have to get from the source;
 * 	do it uniformly at random
 */

int updateTreeFindMissingBlock(UpdateTree * ut,context *c){
	return updateTreeFindMissingBlockRange(ut,c,0,c->numberOfBlocks-1);
}

int updateTreeFindMissingBlockRange(UpdateTree * ut,context *c,int start, int stop){
	int i,j,off;

	if(ut->root == NULL)	// no neighbors yet :(
		return -1;
	pthread_mutex_lock(ut->lock);
	assert(stop>=stop);
	assert(stop<c->numberOfBlocks);
	off = abs(mrand48())%(stop-start+1);
	/*
	i=start+off;		// pick a random starting point
	do{
		if((!blocksIsSet(ut->root->data,i))&&(!blocksIsSet(c->dlBitMap,i))){
			pthread_mutex_unlock(ut->lock);
			return i;
		}
		i++;
		if(i>=c->numberOfBlocks)
			i=0;
	} while(i!=start);
	*/

	for(i=0;i<(stop-start+1);i++){
		j=(start+off+i)%(stop-start+1);	// loop within start:stop range from random middle point
		if((!blocksIsSet(ut->root->data,j))&&(!blocksIsSet(c->dlBitMap,j))){
			if((c->downloadType==DOWNLOADTYPE_ADAPTIVE)
					&&(blocksIsSet(c->ac->inProgress,j)
					&&(c->phase2==0)))
				continue;
			pthread_mutex_unlock(ut->lock);
			return j;
		}
	}

	pthread_mutex_unlock(ut->lock);
	return -1;	// all blocks are in the updateTree
}
	

/* int updateTreeExists(UpdateTree *ut,int block);
 * 	is this block in the updateTree?
 */
int updateTreeExists(UpdateTree *ut,int block){
	assert(ut);
	if(ut->root)
		return blocksIsSet(ut->root->data,block);
	return 0;
}

int updateTreeRevoke(UpdateTree* ut, unsigned int ip, unsigned short port) {
  UpdateNode *un;
	pthread_mutex_lock(ut->lock);
	un = UpdateNodeLookup(ut, ip, port);
	if (un != NULL) {
		Dprintf("updateTreeRevoke\n");
		blocksClear(un->data);
		un->revoked = 1;
	}
	updateTreeProcessUpdate(ut);
	pthread_mutex_unlock(ut->lock);
	return 0;
}
/* int printStatus(struct context *);
 * 	print how many updates we have, and how much we have downloaded
 */

int printStatus(context *c){
	int done;
	int count,knownblocks;
	assert(c);
	if(c->downloadType==DOWNLOADTYPE_ADAPTIVE){
		if(DEBUGLEVEL)
			return printAdaptiveStatus(c);
		else
			return printAdaptiveStatusLine(c);
	}
	pthread_mutex_lock(c->updateTree->lock);
	done = blocksCountBits(c->dlBitMap);
	knownblocks=(c->updateTree->root ==NULL)?0:blocksCountBits(c->updateTree->root->data);
	count=updateTreeCountNodes(c);
	Dprintf("STATUS: %d (%.3f%%) blocks in updateTree (%d clients), %d downloaded(%.3f%%)\n",
			knownblocks,
			100*(float)knownblocks/c->numberOfBlocks,
			count,
			done, 100*(float)done/c->numberOfBlocks);
	pthread_mutex_unlock(c->updateTree->lock);

	return 0;
}

int updateTreeCountNodes(context *c){
	int count=0,i;
	for(i=0;i<c->updateTree->numLeaves;i++){
		if(!c->updateTree->leaves[i]->revoked)
			count++;
	}
	return count;
}


/* int printAdaptiveStatus(context *c)
 * 	print useful info about our adaptive dl
 */

int printAdaptiveStatus(context *c){
	int done,count,knownblocks;
	static int wasDone=0;
	int nSources=0, nPeers=0;
	int i;
	time_t now;
	assert(c);
	if((c->ac==NULL)||(DEBUGLEVEL<1))
		return 0;
	c->ac->nConnections=c->ac->nConnections;	// this is just to shut up valgrind for uninitted mem
	count=updateTreeCountNodes(c);
	pthread_mutex_lock(c->updateTree->lock);
	done = blocksCountBits(c->dlBitMap);
	knownblocks=(c->updateTree->root ==NULL)?0:blocksCountBits(c->updateTree->root->data);
	now = time(NULL);
	for(i=0;i<c->ac->nConnections;i++)
		if(c->ac->connection[i]->source)
			nSources++;
		else
			nPeers++;
	Dprintf("STATUS: %d (%.3f%%) in updateTree (%d nodes)"
			", %d dl;d(%.3f%%) %d ; "
			"%d %d s/p conns; "
			"nUpdates %d "
			"nNeighbors %d: %ld %ld %s %d\n",
			knownblocks,
			100*(float) knownblocks/c->numberOfBlocks,
			count,
			done, 100*(float)done/c->numberOfBlocks, done-wasDone,
			nSources,nPeers,
			c->desiredNumOfUpdates,
			c->neighborMgr->nNeighbors,
			now, now- c->startTime,
			c->backoffDidBackOff?"B":"C", c->phase2);
	wasDone=done;
	pthread_mutex_unlock(c->updateTree->lock);
	if(DEBUGLEVEL>=1){
		pthread_mutex_lock(c->neighborMgr->lock);
		fprintf(stderr,"NM_LINKS %ld %d:%u: ", now, c->localIp, neighborManagerGetPort(c->neighborMgr));
		for(i=0;i<c->neighborMgr->nNeighbors;i++){
			fprintf(stderr,"%d:%u ",c->neighborMgr->neighbors[i]->ip,
					c->neighborMgr->neighbors[i]->remote_nm_port);
		}
		fprintf(stderr,"\n");
		pthread_mutex_unlock(c->neighborMgr->lock);
	}
	if((done == c->numberOfBlocks)&&(!blocksFull(c->dlBitMap)))
		Dprintf("USELESS FUCKING BUGS\n");

	return 0;
}

int printAdaptiveStatusLine(context *c){
	int done,count,knownblocks;
	static int wasDone=0;
	int nSources=0, nPeers=0;
	int i;
	time_t now;
	static time_t then=0;
	double rate;
	char * units="B";
	assert(c);
	if((c->ac==NULL)||(c->quiet))
		return 0;
	c->ac->nConnections=c->ac->nConnections;	// this is just to shut up valgrind for uninitted mem
	count=updateTreeCountNodes(c);
	pthread_mutex_lock(c->updateTree->lock);
	done = blocksCountBits(c->dlBitMap);
	knownblocks=(c->updateTree->root ==NULL)?0:blocksCountBits(c->updateTree->root->data);
	now = time(NULL);
	if(then==0)
		then=c->startTime;
	if(now-then>0)
		rate = c->blockSize*(double)(done-wasDone)/(now-then);
	else 
		rate=0;
	if(rate>1024){
		units="KB";
		rate/=1024;
	}
	if(rate>1024){
		units="MB";
		rate/=1024;
	}
	if(rate>1024){
		units="GB";
		rate/=1024;
	}
	for(i=0;i<c->ac->nConnections;i++)
		if(c->ac->connection[i]->source)
			nSources++;
		else
			nPeers++;
	printf("%.3f%% Done (%d/%d blocks) %4.3f %s/s --  %.3f%% Discovered (%d %s). "
			"S/P/N=%d,%d,%d: FLAGS=%s\r",
			100*(float) done/c->numberOfBlocks,
			done,
			c->numberOfBlocks,
			rate,units,
			100*(float)done/c->numberOfBlocks, 
			count,
			count==1?"node":"nodes",
			nSources,nPeers,
			c->neighborMgr->nNeighbors,
			c->backoffDidBackOff?"B":"C");
	fflush(stdout);
	wasDone=done;
	pthread_mutex_unlock(c->updateTree->lock);
	return 0;
}
/* int updateTreeFirstNeedBlock(struct context * ctx,unsigned int ip,unsigned short port,int start);
 * 	given a specific node, and a starting block, find the next block that that
 * 	node has that we don't already have; else return -1
 */

int updateTreeFirstNeedBlock(struct context * ctx,unsigned int ip,
		unsigned short port,int start){
	int block,i;
	blocks * data;
	pthread_mutex_lock(ctx->updateTree->lock);
	for(i=0;i<ctx->updateTree->numLeaves;i++){
		if((ctx->updateTree->leaves[i]->ip == ip )&&(
				ctx->updateTree->leaves[i]->port==port))
			break;
	}
	if(i>=ctx->updateTree->numLeaves){ // didn't find the node!
		pthread_mutex_unlock(ctx->updateTree->lock);
		Dprintf("updateTreeFirstNeedBlock:: Couldn't find who we were talking to\n");
		return -1;
	}
	data = ctx->updateTree->leaves[i]->data;
	for(i=start;i<ctx->numberOfBlocks;i++){
		if((blocksIsSet(data,i))&&(!blocksIsSet(ctx->dlBitMap,i))){
			if((ctx->downloadType==DOWNLOADTYPE_ADAPTIVE)
					&& (ctx->phase2==0)
					&& (blocksIsSet(ctx->ac->inProgress,i)))
				continue;
			break;
		}
	}
	if(i<ctx->numberOfBlocks)
		block=i;
	else 
		block=-1;
	pthread_mutex_unlock(ctx->updateTree->lock);
	return block;
}


int updateTreeMarkLeafNotBusy(struct context *c, connection_t* con){
  unsigned int addr = connection_get_peer_ip(con);
  unsigned short port = connection_get_peer_port(con);
  UpdateTree* ut = c->updateTree;
  UpdateNode* node;
  int x;

  Dprintf2(2, "updateTreeMarkLeafNotBusy called\n");
  pthread_mutex_lock(ut->lock);
  for(x=0; x<ut->numLeaves; x++) {
    node = ut->leaves[x];
    if(node->ip == addr && node->port == port) {
      if(node->busy == 0) {
	Dprintf2(2, "UPDATE TREE: Node marked as not busy twice\n");
      }
      node->busy = 0;
      pthread_mutex_unlock(ut->lock);
      return 0;
    }
  }
  Dprintf2(2, "UPDATE TREE::markNotBusy Could not find the addr in tree?\n");
  pthread_mutex_unlock(ut->lock);
  return 1;
}
int updateTreeMarkLeafBusy(struct context *c, download *dl){
  UpdateTree* ut = c->updateTree;
  UpdateNode* node;
  int x;

  Dprintf2(2, "updateTreeMarkLeafBusy called\n");
  pthread_mutex_lock(ut->lock);
  for(x=0; x<ut->numLeaves; x++) {
    node = ut->leaves[x];
    if(node->ip == dl->ip && node->port == dl->port) {
      if(node->busy == 1)
	Dprintf2(2, "UPDATE TREE: Node marked as busy twice\n");
      node->busy = 1;
      pthread_mutex_unlock(ut->lock);
      return 0;
    }
  }
  Dprintf2(2, "UPDATE TREE::markBusy Could not find addr in tree?\n");
  pthread_mutex_lock(ut->lock);
  return 1;
}





/************************************************************************
 * 	unittesting stuff
 ************************************************************************/

int do_updateTree_tests(){
	int blockSize,totalSize;
	const int dbuflen=16;
	UpdateTree * ut;
	context *c;
	blocks *b1, *b2,*b3;
	download * dbuf[dbuflen];
	int len = dbuflen;
	int tmp;

	c = contextCreate();
	printf("-------------- testing updateTree.o \n");
	blockSize =16;
	totalSize=128;		// should make 8 blocks
	c->blockSize=blockSize;
	c->fileSize = totalSize;
	c->numberOfBlocks=totalSize/blockSize;

	ut = updateTreeNew(10);		// ten leaves
	b1 = blocksNew(blockSize,totalSize);
	b2 = blocksNew(blockSize,totalSize);
	b3 = blocksNew(blockSize,totalSize);
	blocksSet(b1,0);
	blocksSet(b1,1);
	blocksSet(b2,1);
	blocksSet(b2,3);
	blocksSet(b3,4);
	blocksSet(b3,5);

	updateTreeUpdate(c,ut,0,0,b1);
	updateTreeUpdate(c,ut,1,1,b2);
	updateTreeUpdate(c,ut,2,2,b3);

	tmp = updateTreeSearch(ut,0, &len,dbuf);
	assert(tmp==1);
	assert(tmp==len);
	assert(dbuf[0]->ip == 0);
	len = dbuflen;
	tmp = updateTreeSearch(ut,1, &len,dbuf);
	assert(tmp==2);
	assert((dbuf[0]->ip == 0)||(dbuf[0]->ip == 1));
	assert((dbuf[1]->ip == 0)||(dbuf[1]->ip == 1));
	assert(tmp==len);

	tmp = updateTreeFindMissingBlock(ut,c);
	assert(tmp!=0);
	assert(tmp!=1);
	assert(tmp!=3);
	assert(tmp!=4);
	assert(tmp!=5);
	printf("Got missing block %d : not 0,1,3,4,5\n",tmp);

	return 0;
}
