#include <stdio.h>
#include <assert.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <stdlib.h>
#include <string.h>


#include "connection.h"
#include "updates.h"
#include "utils.h"
#include "updateTree.h"

/***********************************************************************
 * private functions
 ************************************************************************/

// do for each connection
int getUpdates(neighbor*, struct context *);
int sendUpdates(neighbor*, struct context *);
// do for each single update
int sendSingleUpdate(int fd, update *,struct context *);
int getSingleUpdate(int fd, update *,context *);
/* block reading/writing rules */
int doReadBitmapUpdate(int fd,update *u,context *);
int doWriteBitmapUpdate(int fd,update *u,context *);
void * sendAllUpdates(void * contextPtr);

int addOurUpdates(context *c);




/* pthread_t * spawnSendUpdate(context *):
 * 	start the sendUpdates thread and return it
 */

pthread_t * spawnSendUpdate(context *c){
  pthread_t *thread;
  int err;

  thread = malloc(sizeof(pthread_t));
  err=pthread_create(thread,NULL,sendAllUpdates,c);
  /*
    pthread_detach(*thread);	// make it so we don't have to clean
    // this thread up
    */
  if(err) 
    return NULL;
  else
    return thread;
}

/* int getAllUpdates(context *, neighbor **list, int listlen);
 * 	go to each connection listed and get the updates from
 * 		it, and process it into the updateTree
 * 		ASSUMES nm is locked
 */

int getAllUpdates(context *c, neighbor **list, int listlen){
  int i,err;
  char buf[BUFLEN];
  unsigned short port;

  //Dprintf("Doing getAllUpdates:: %d peers\n",listlen);
  strncpy(buf,"DISCONNECTED",BUFLEN);
  for(i=0;i<listlen;i++){
    err = getUpdates(list[i],c);
    if(err){
      connection_get_peer_addr(list[i]->link,buf,BUFLEN);
      port = connection_get_peer_port(list[i]->link);
      Dprintf("Bad update (err=%d) from: %s:%u, <<< WILL BE REMOVED >>>!\n", err, buf, port);
      list[i]->del_flag = 1;
    }
    neighborManagerRemovePeer(c->neighborMgr, list[i]);
  }

  return 0;
}

/* void * sendAllUpdates(void * contextPtr);
 *	step through each neighbor, write some updates to them,
 *	and continue; remove them from the list if there's an error
 */

void * sendAllUpdates(void * contextPtr){
  context * c = contextPtr;
  int i;
  int err;
  char buf[BUFLEN]="Unknown";
  unsigned short port=-1;
  time_t start,stop;
  neighborManager *nm;

  nm = c->neighborMgr;
  while(1){
		// Add our updates is now at source_get
    addOurUpdates(c);	// submit our own updates
    start=time(NULL);
    Dprintf2(4,"Doing sendAllUpdates:: %d peers\n",nm->nNeighbors);
    for(i=0;i<nm->nNeighbors;i++){
      if(nm->neighbors[i]->del_flag)
	neighborManagerRemovePeer(c->neighborMgr, nm->neighbors[i]);
      else {
	err=sendUpdates(nm->neighbors[i],c);
	if(err){
	  connection_get_peer_addr(nm->neighbors[i]->link,buf,BUFLEN);
	  port = connection_get_peer_port(nm->neighbors[i]->link);
	  Dprintf("Error sending update to %s:%u\n",buf,port);
	  //neighborDestroy(list[i]);
	  neighborManagerRemovePeer(c->neighborMgr,nm->neighbors[i]);
	}
      }
    }
    stop = time(NULL);	// cheese timing
    if(start-stop<UPDATES_TIME_BETWEEN_UPDATES)
      sleep(UPDATES_TIME_BETWEEN_UPDATES+stop-start);
    printStatus(c);
  }

  return NULL;

}


/* int getUpdates(int fd, context *);
 * 	get all of the updates from a single socket
 *	and commit them to the updatetree in the passed context
 */
int getUpdates(neighbor *n, context *c){
  char buf[BUFLEN];
  int err,status;
  int count,desired, type;	// "desired" is the num of updates they want from us
  int i;
  update u;
  int fd;
  int remote_nm_port;

  assert(n);
  fd = neighborGetFD(n);

  err = timeout_read_line(&status,fd,buf,BUFLEN,UPDATES_TIMEOUT);
  if(err) {
    return err;	// timeout
  }
  if(status<0) {
    return status;	//read error
  }
  if(status==0) {
    return -1;	//end of file
  }
  err = !sscanf(buf,UPDATESTR" %d %d %d %d",&type,&desired,&count,&remote_nm_port);
  if(err){
    Dprintf("Got BAD update '%s'\n",buf);
    return -2;
  }
  if(type != c->blockRule){
    Dprintf("Got UPDATE for BlockRule %d, not %d\n",type,c->blockRule);
    return -3;
  }
  if(count<1){
    Dprintf("Got BAD count %d\n", count);
    return -4;
  }
  n->updatesPerRound=desired;	// how many updates they want from us on their next round
  n->remote_nm_port=remote_nm_port;	// tells us who they really are (ip:port pair)
  u.type = type;
  for(i=0;i<count;i++){
    err=getSingleUpdate(fd, &u,c);
    if(err) {
      Dprintf("getUpdates: getSingleUpdate() did not succeed.\n");
      return -5;
    }
    updateTreeUpdate(c,c->updateTree,u.ip,u.port,u.data);
  }

  return 0;
}

/* int sendUpdates(neighbor*, context *);
 * 	send some number of updates to this neighbor, using
 * 	the numbers in the context
 */
int sendUpdates(neighbor* n, context *c){
  char buf[BUFLEN];
  update * ulist[MAXNUMOFUPDATES];
  int count,fd;
  int err, status,i;

  count = neighborGetUpdates(n);
  count = getRandomUpdates(c->updateTree,c, ulist,count);
  snprintf(buf,BUFLEN,"%s %d %d %d %d\n",
	   UPDATESTR,
	   c->blockRule,
	   c->desiredNumOfUpdates,
	   count,
	   neighborManagerGetPort(c->neighborMgr));
  // Dprintf("sendUpdate: %s\n", buf);
  fd = neighborGetFD(n);
  err = timeout_writeall(&status,fd,buf,strlen(buf),UPDATES_TIMEOUT);
  if(err)
    return err;
  if(status<strlen(buf))
    return -1;
  for(i=0;i<count;i++){
    err = sendSingleUpdate(fd,ulist[i],c);
    free(ulist[i]);
    if(err)
      return -2;
  }
  return 0;
}


/* int getSingleUpdate(int fd, update *);
 * 	read the ip/port and then demux on the type, and issue a 	
 * 	rule specific block get
 */

int getSingleUpdate(int fd, update *u, context *c){
  int err, status;
  //char ipbuf[BUFLEN];

  err = timeout_readall(&status, fd, &u->ip,sizeof(u->ip),UPDATES_TIMEOUT);
  if((err) || (status != sizeof(u->ip))) {
    Dprintf("getSingleUpdate: timeout_read_all(ip) error.\n");
    return -1;
  }
  u->ip = ntohl(u->ip);
  err = timeout_readall(&status, fd, &u->port,sizeof(u->port),UPDATES_TIMEOUT);
  if((err) || (status != sizeof(u->port))) {
    Dprintf("getSingleUpdate: timeout_read_all(port) error.\n");
    return -1;
  }
  u->port = ntohs(u->port);
	// Dprintf("Read Update from: %s:%d\n", inet_ntop(AF_INET,&u->ip,ipbuf,BUFLEN), u->port);
  switch(u->type){
  case BLOCKRULE_BITMAP:
    return doReadBitmapUpdate(fd,u,c);
  default: 
    fprintf(stderr,"NON-EXISTENT BLOCK RULE: %d\n",u->type);
    abort();
  };

  return 0;	// should never reach here
}

/* int sendSingleUpdate(int fd, update *,context *);
 * 	send the ip/port and then demux on the type and
 * 	issue a rule specific block write
 */

int sendSingleUpdate(int fd, update *u,context *c){
  int err, status;
  unsigned int ip;
  unsigned short port;

  ip = htonl(u->ip);
  port = htons(u->port);

  err = timeout_write(&status, fd, &ip,sizeof(ip),UPDATES_TIMEOUT);
  if((err) || (status != sizeof(ip)))
    return -1;
  err = timeout_write(&status, fd, &port,sizeof(port),UPDATES_TIMEOUT);
  if((err) || (status != sizeof(port)))
    return -1;
  //  Dprintf("Wrote: %d:%d\n", u->ip, u->port);
  switch(u->type){
  case BLOCKRULE_BITMAP:
    return doWriteBitmapUpdate(fd,u,c);
  default:
    fprintf(stderr,"NON-EXISTENT BLOCK RULE: %d\n",u->type);
    abort();
  };
  return 0;	// should never get here
}


/* int doReadSizeUpdate(int fd,updates *u);
 * 	read a data chunk with rule 0, and turn it into a block
 */
int doReadBitmapUpdate(int fd,update *u, context *c){
  assert(u);
  // don't free the existing block, but it is in use in the tree
  u->data=blocksNew(c->blockSize,c->fileSize);
  return blocksRead(u->data,fd,UPDATES_TIMEOUT);
}

/* int doWriteBitmapUpdate(int fd,update *u,context *);
 * 	write an Bitmap update to the fd 
 */

int doWriteBitmapUpdate(int fd,update *u,context *c ){
  assert(u);
  assert(c);
  return blocksWrite(u->data,fd,UPDATES_TIMEOUT);
}

/* int calcBlocks(struct context *c);
 * 	calculate the blocks breakdown based on the blocking rule
 */

int calcBlocks(struct context *c){
  switch(c->blockRule){
  case 0:
	  if(c->blockingPriority ==BLOCKSIZEPRIORITY){
		c->numberOfBlocks=c->fileSize/c->blockSize;
		if(c->fileSize%c->blockSize)
			c->numberOfBlocks++;
	  }else {
		assert(c->blockingPriority == NUMBLOCKSPRIORITY);
		// let's see if we can get this right
		  c->blockSize = c->fileSize/c->numberOfBlocks;
		if( c->fileSize % c->numberOfBlocks)
			c->blockSize++;
	  }

    break;
  default:
    fprintf(stderr,"Unknown blocking rule %d\n",c->blockRule);
    abort;
  }
  Dprintf("Getting %d blocks of size %d for a file of size %d\n",
		  c->numberOfBlocks, c->blockSize, c->fileSize);
  return 0;
}


/* int addOurUpdates(c);
 * 	put our own downloaded stuff into the updatesTree
 */

int addOurUpdates(context *c){
  updateTreeUpdate(c,c->updateTree,c->localIp, neighborManagerGetPort(c->neighborMgr), 
		   blocksDuplicate(c->dlBitMap));
  return 0;
}


