#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <time.h>

#include "hashkeys.h"
#include "bf.h"

void SetBit(UINT8 *theData, UINT32 thePos) 
{
  UINT32 bytePos = thePos/8;
  UINT32 bitPos = thePos%8;

  theData[bytePos] |= 1 << bitPos;
}

void ResetBit(UINT8 *theData, UINT32 thePos) 
{
  UINT32 bytePos = thePos/8;
  UINT32 bitPos = thePos%8;

  theData[bytePos] &= ~(1 << bitPos);
}


UINT8 GetBit(UINT8 *theData, UINT32 thePos) 
{
  UINT8 result;
  UINT32 bytePos = thePos/8;
  UINT32 bitPos = thePos%8;

  result = (theData[bytePos] >> bitPos) & 0x01;
  return result;
}

void BFInit(BFCTX *theCTX, 
	    UINT32 theSize, 
	    UINT8 theFuncNum) 
{
  UINT32 i;
  UINT8  funcBits=0;
  size_t sizeBytes;

  sizeBytes = theSize/8;
  if ((theSize%8) != 0) {
    // should add one byte because / is the floor operator
    ++sizeBytes;
  }

  theCTX->sizeBits = theSize;
  theCTX->sizeBytes = sizeBytes;

  // init hash func CTX
  for (i = 0; i < 128; ++i) {
    if (theSize == (1 << i)) {
      funcBits = i;
      break;
    }
    else if (theSize < (1 << i)) {
      assert(theSize == (1 << i));
    }
  }
  HashKeysInit(&(theCTX->hashCTX), funcBits, theFuncNum);

  // init BF buffer
  theCTX->data = malloc(sizeBytes);
  for (i = 0; i < sizeBytes; ++i) {
    theCTX->data[i] = 0;
  }
}


void BFFree(BFCTX *theCTX) 
{
  if (theCTX->data != NULL) {
    free(theCTX->data);
  }
}


void BFSet(BFCTX *theCTX, 
	   UINT32 thePos)
{
  UINT32 i;

  // preconditions
  assert(theCTX->data != NULL);

  // calculate hash function values for thePos
  HashKeysCalcInt(&(theCTX->hashCTX), thePos);

  // set corresponding bits
  for (i = 0; i < (theCTX->hashCTX).funcNum; ++i) {
    SetBit(theCTX->data, (theCTX->hashCTX).keys[i]);
  }
}


void BFReset(BFCTX *theCTX, 
	     UINT32 thePos)
{
  UINT32 i;

  // preconditions
  assert(theCTX->data != NULL);

  // calculate hash function values for thePos
  HashKeysCalcInt(&(theCTX->hashCTX), thePos);

  // set corresponding bits
  for (i = 0; i < (theCTX->hashCTX).funcNum; ++i) {
    ResetBit(theCTX->data, (theCTX->hashCTX).keys[i]);
  }
}


int BFIsSet(BFCTX *theCTX, 
	    UINT32 thePos) 
{
  UINT32 i;

  // preconditions
  assert(theCTX->data != NULL);

  // calculate hash function values for thePos
  HashKeysCalcInt(&(theCTX->hashCTX), thePos);

  // check corresponding bits
  for (i = 0; i < (theCTX->hashCTX).funcNum; ++i) {
    if (GetBit(theCTX->data, (theCTX->hashCTX).keys[i]) == 0) {
      return 0;
    }
  }

  return 1;
}


UINT8 BFGetOptimalFuncNum(double theMNRatio)
{
  UINT8 result;
  double tempResult = log(2) * theMNRatio;

  result = ceil(tempResult);

  return result;
}


void DoBFTest()
{
  UINT32 i;
  UINT32 BF_SIZE = 128;
  UINT32 UNIVERSE_SIZE = 1024;
  double INSERT_P = 0.016;
  double randomNum;
  UINT8  *bitMap;
  UINT32 falsePosNum = 0;
  UINT32 falseNegNum = 0;
  UINT32 setSize = 0;

  BFCTX bfCtx;

  // initialize random number generator
  srand((unsigned int)time(NULL));

  // initialize BF and BitMap
  BFInit(&bfCtx, BF_SIZE, BFGetOptimalFuncNum(8));
  bitMap = malloc(UNIVERSE_SIZE/8);
  for (i = 0; i < UNIVERSE_SIZE/8; ++i) {
    bitMap[i] = 0;
  }

  // insert random subset
  for (i = 0; i < UNIVERSE_SIZE; ++i) {
    randomNum = (double)rand() / (double)RAND_MAX;
    if ((randomNum < INSERT_P) && setSize < 16) {
      SetBit(bitMap, i);
      BFSet(&bfCtx, i);
      ++setSize;
    }
  }

  // check for false positives and false negatives
  for (i = 0; i < UNIVERSE_SIZE; ++i) {
    if ((GetBit(bitMap, i) == 1) &&
	(BFIsSet(&bfCtx, i) == 0)) {
      ++falseNegNum;
    }
    else if ((GetBit(bitMap, i) == 0) &&
	     (BFIsSet(&bfCtx, i) == 1)) {
      ++falsePosNum;
    }
  }

  printf("false positives: %ld / %ld\n", falsePosNum, setSize);
  printf("false negatives: %ld / %ld\n", falseNegNum, setSize); 

  BFFree(&bfCtx);
}
