#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <assert.h>

#include <utils.h>
#include <connection.h>
#include <protocol.h>


static const unsigned short HANDSHAKE_BSIZE = 1024;
const char* MAGIC = "SLURPIE";

// Private Helpers
int connection_handshake(connection_t* c);	// inititate handshake
int connection_gethandshake(connection_t* c);	// negotitate handshake

// Creates a connection structure and puts the inner socket on the listening state
connection_t* connection_create(protocol_t proto, version_t ver) {
	connection_t* rv = (connection_t*)malloc(sizeof(connection_t));

	if (rv != NULL)  {
		rv->sock = -1;
		rv->timeout = 10000;
		rv->err = 0;

		rv->proto = proto;
		rv->ver = ver;
	}
	else 
		printf("connection_create: Warning could not allocate %d bytes", sizeof(connection_t));

	return rv;
}

// Calls shutdown on the socket and deletes the cp
void connection_destroy(connection_t* cp) {
	if (cp) {
		if (cp->sock > 0){
			shutdown(cp->sock,2);
			close(cp->sock);
		}
		free(cp);
	}
}

int connection_listen(connection_t* c, unsigned short port) {
	int rv = 0;
	c->sock = make_tcp_server(&port);
	if (!(c->sock > 0)) {
		c->err = c->sock;
		c->sock = -1;
		rv = -1;
	}
	return rv;
}

connection_t* connection_accept(connection_t* lp) {
	connection_t* neo = NULL;
	struct sockaddr_in neo_addr;
	int try_again = 0;
#ifdef __darwin__
	int neo_len;
#else
	socklen_t neo_len;
#endif
	do {
		neo = connection_create(lp->proto, lp->ver);
		assert(neo);
		neo_len = sizeof(neo_addr);
		
		neo->sock = accept(lp->sock, (struct sockaddr*)&neo_addr, &neo_len);
		if (neo->sock == -1) {
			lp->err = errno;
			connection_destroy(neo);
			neo = NULL;
		}
		else {
			if ((lp->err = connection_gethandshake(neo)) != 0) {
				connection_destroy(neo);
				neo = NULL;
				try_again = 1;
			}
		}
	}while(try_again);
	return neo;
}

int connection_connect(connection_t* neo, const char* addr, unsigned short port) {
	return connection_connect_from_port(neo,addr,port,0);
}
int connection_connect_from_port(connection_t* neo, const char* addr, unsigned short port,unsigned short sport) {
	int rv = 0;
	if ((neo->sock = make_tcp_connection_from_port(addr, port,sport)) > 0) {
		if ((neo->err = connection_handshake(neo)) != 0) {
			rv = -1;
		}
	}
	else {
		neo->err = errno;
		perror("connection_connect_from_port::make_tcp_connection_from_port");
		rv = -1;
	}
	return rv;
}
int connection_is_valid(connection_t* s) {
	return s->sock > 0 && s->err == 0;
}

const char* connection_get_peer_addr(connection_t* t, char* buf, int len) {{{
	const char* rv = NULL;
	struct sockaddr_in sin;
#ifdef __darwin__
	int sin_len;
#else
	socklen_t sin_len;
#endif
	sin_len = sizeof(sin);
	if (getpeername(t->sock, (struct sockaddr *)&sin, &sin_len) != 0) 
		t->err = errno;
	else 
		rv = inet_ntop(sin.sin_family, &sin.sin_addr, buf, len);
	return rv;
}}}
unsigned short connection_get_peer_port(connection_t* t) {{{
	unsigned short rv =0;
	struct sockaddr_in sin;
#ifdef __darwin__
	int len;
#else
	socklen_t len;
#endif 
	len = sizeof(sin);

	if (getpeername(t->sock, (struct sockaddr *)&sin, &len) != 0) 
		t->err = errno;
	else 
		rv = ntohs(sin.sin_port);

	return rv;
}}}
const char* connection_get_sock_addr(connection_t* t, char* buf, int len) {{{
	const char* rv = NULL;
	struct sockaddr_in sin;
	unsigned int tmp;
#ifdef __darwin__
	int sin_len;
#else
	socklen_t sin_len;
#endif
	sin_len = sizeof(sin);
	if (getsockname(t->sock, (struct sockaddr *)&sin, &sin_len) != 0) 
		t->err = errno;
	else {
		rv = inet_ntop(sin.sin_family, &sin.sin_addr, buf, len);
		if (strcmp(rv, "0.0.0.0") == 0) {
			// replaced with my equiv function, cuz I'm anal :(
			//char** myaddrs = my_addrs(&addrtype);
			//rv = inet_ntop(addrtype, myaddrs[0], buf, len);
			tmp = getLocalHostIP();
			rv = inet_ntop(AF_INET, &tmp, buf, len);

		}
	}
	return rv;
}}}
/* unsigned int connection_get_peer_ip(connection_t *c);
 * 	return the ip of the connection in some sort of useful form
 * 		different from connection_get_sock_addr() as this is a useful
 * 		function, and that is not
 */
unsigned int connection_get_peer_ip(connection_t *c){
	struct sockaddr_in sin;
	int sin_len;
	unsigned int ip;

	sin_len = sizeof(sin);
	if (getpeername(c->sock, (struct sockaddr *)&sin, &sin_len) != 0){
			c->err = errno;
			return 0;
	}
	ip = sin.sin_addr.s_addr;
	return ip;
}

unsigned short connection_get_sock_port(connection_t* t) {{{
	unsigned short rv =0;
	struct sockaddr_in sin;
#ifdef __darwin__
	int len;
#else
	socklen_t len;
#endif 
	len = sizeof(sin);

	if (getsockname(t->sock, (struct sockaddr *)&sin, &len) != 0) 
		t->err = errno;
	else 
		rv = ntohs(sin.sin_port);

	return rv;
}}}
int connection_get_command_and_parameters(connection_t* c, char* command, size_t len) {
	int rv = 0;
	char token[BUFLEN];
	const char* proto = proto_enum_to_string(c->proto);
	int status;

	// Get magic
	rv = (rv)? rv :timeout_read_token(&status, c->sock, token, BUFLEN, c->timeout);
	rv = (rv)? rv :strcmp(token, MAGIC);

	// Get proto
	rv = (rv)? rv :timeout_read_token(&status, c->sock, token, BUFLEN, c->timeout);
	rv = (rv)? rv :strcmp(token, proto);

	// Snarf til eol
	rv = (rv)? rv :timeout_read_line(&status, c->sock, command, len, c->timeout);

	return rv;
}

// Private Helpers
int connection_handshake(connection_t* c) {{{
	int rv = 0;
	if ( c->proto != PROTO_UNKNOWN ) {
		char ch;
		const char* proto = proto_enum_to_string(c->proto);
		const char* command = "VERSION";
		char handshakeBuf[HANDSHAKE_BSIZE];
		char str[3][HANDSHAKE_BSIZE];
		version_t rem_ver;
		int len;
		int itm_read;
		int status;

		// Builds our version line
		len = sprintf(handshakeBuf, "%s %s %s %u.%u\n", MAGIC, proto, command, c->ver.major, c->ver.minor);

		// Sends our version line
		rv = (rv) ? rv : timeout_writeall(&status, c->sock, handshakeBuf, len, c->timeout);
		// If we did not timeout, read the handshake line from the other side
		rv = (rv) ? rv : (timeout_read_line(&status, c->sock, handshakeBuf, sizeof(handshakeBuf), c->timeout));
		// If we did not timeout again, parse the handshake line.
		rv = (rv)? rv : !(
				(itm_read = sscanf(handshakeBuf, "%s %s %s %u.%u%c", str[0], str[1], str[2], &rem_ver.major, &rem_ver.minor, &ch)) == 5 
				&& strcmp(str[0], MAGIC) == 0 
				&& c->proto == proto_string_to_enum(str[1])
				&& strcmp(str[2], command) == 0);

		if (rv == 0) {
			c->ver = (version_lt(&c->ver, &rem_ver))? c->ver: rem_ver;
			//Dprintf("Handshake succeeded!\n");
		}
		else {
			Dprintf("Handshake failed! rv=%d, proto=%d\n",rv,
				c->proto);
		}
	}
	return rv;
}}}


// Private Helpers
int connection_gethandshake(connection_t* c) {{{
	int rv = 0;
	if ( c->proto != PROTO_UNKNOWN ) {
		char ch;
		const char* proto = proto_enum_to_string(c->proto);
		const char* command = "VERSION";
		char handshakeBuf[HANDSHAKE_BSIZE];
		char str[3][HANDSHAKE_BSIZE];
		version_t rem_ver;
		int len;
		int itm_read;
		int status;
		
		// Read what they want to connect as
		rv = (rv) ? rv : (timeout_read_line(&status, c->sock, handshakeBuf, sizeof(handshakeBuf), c->timeout));
		// If we did not timeout, parse the handshake line.
		rv = (rv)? rv : !(
				(itm_read = sscanf(handshakeBuf, "%s %s %s %u.%u%c", str[0], str[1], str[2], &rem_ver.major, &rem_ver.minor, &ch)) == 5 
				&& strcmp(str[0], MAGIC) == 0 
				&& strcmp(str[2], command) == 0);
		// this sets the protocol; if it is unknown, it sets it to UNKNOWN
		rv = (rv)? rv : !(c->proto = proto_string_to_enum(str[1]));	// set the protocol
		proto = proto_enum_to_string(c->proto);

		// Builds our version line
		len = sprintf(handshakeBuf, "%s %s %s %u.%u\n", MAGIC, proto, command, c->ver.major, c->ver.minor);;

		// Sends our version line
		rv = (rv) ? rv : timeout_writeall(&status, c->sock, handshakeBuf, len, c->timeout);

		if (rv == 0) {
			c->ver = (version_lt(&c->ver, &rem_ver))? c->ver: rem_ver;
			// Dprintf("Get Handshake succeeded!\n");
		}
		else {
			Dprintf("Handshake failed!: rv=%d\n",rv);
		}
	}
	return rv;
}}}

// return the file descriptor so we can select() on it
int connect_get_fd(connection_t* c){
	assert(c);
	return c->sock;
}

