#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <pthread.h>

#include "pmdb.h"
#include "hash.h"

/* drinkersList *drinkersListNew(int maxLength){
 * 	make an empty drinkersList struct
 */
drinkersList *drinkersListNew(char *key,int maxLength){
	drinkersList * ret;

	ret = malloc(sizeof(drinkersList));
	if(!ret){
		perror("drinkersListNew::malloc");
		abort();
	}
	ret->length =0;
	ret->lastIndex=0;
	ret->maxLength=maxLength;
	ret->data = malloc(sizeof(ip_port_t*)*maxLength);
	ret->key = strdup(key);
	ret->next = NULL;
	if(!ret->data){
		perror("drinkersListNew::malloc");
		abort();
	}
	return ret;
}

/* int drinkersListDestroy(drinkersList *);
 * 	free all of the resources for a drinkersList struct
 */
int drinkersListDestroy(drinkersList * dl){
	int i;
	assert(dl);
	assert(dl->data);
	for(i=0;i<dl->length;dl++){
		assert((dl->data[i]));
		ip_port_destroy(&(dl->data[i]));
	}
	free(dl->data);
	free(dl);
	return 0;
}

/* int drinkersListAdd(drinkersList *dl,ip_port_t *data){
 * 	return 1 if exists, 0 otherwise
 */
int drinkersListExists(drinkersList *dl,ip_port_t *data){
	int i;
	for(i=0;i<dl->length;i++)
		if(ip_port_equal(data,dl->data[i]))
			return 1;
	return 0;
}


/* int drinkersListAdd(drinkersList *dl,ip_port *data);
 * 	add this ip_port to the drinkersList
 * 		- if it is not full, add it to the end
 * 		- if it is, use the lastIndex to bump
 * 		the least recent guy from the list
 */

int drinkersListAdd(drinkersList *dl,ip_port_t *data){
	assert(dl);
	assert(data);
	if(drinkersListExists(dl,data))
		return 1;	// already exists
	if(dl->length<dl->maxLength){
		dl->data[dl->length++]=data;
		return 0;
	}
	// else replace the least recent guy
	dl->data[dl->lastIndex++]=data;
	dl->lastIndex = dl->lastIndex%dl->maxLength;
	return 0;
}

/* pmdb * pmdbNew();
 * 	create a new, empty poor man's database
 */
pmdb * pmdbNew(){
	pmdb * ret;
	int i;

	ret = malloc(sizeof(pmdb));
	if(!ret){
		perror("pmdbNew::malloc");
		assert(ret);
	}

	ret->data = malloc(sizeof(drinkersList*)*PMDB_LENGTH);
	if(!ret->data){
		perror("pmdbNew::malloc");
		assert(ret->data);
	}
	for(i=0;i<PMDB_LENGTH;i++)
		ret->data[i]=NULL;
	ret->lock = malloc(sizeof(pthread_mutex_t));
	if(!ret->lock){
		perror("pmdbNew::malloc");
		assert(ret->lock);
	}
	pthread_mutex_init(ret->lock,NULL);

	return ret;
}

/* int pmdbDestroy(pmdb *);
 * 	free resources in the pmdb
 */

int pmdbDestroy(pmdb *pm){
	int i;
	drinkersList *curr,*prev;
	assert(pm);
	assert(pm->data);
	for(i=0;i<PMDB_LENGTH;i++){
		curr=pm->data[i];
		while(curr){
			prev=curr;
			curr=curr->next;
			drinkersListDestroy(prev);
		}
	}

	free(pm->data);
	free(pm->lock);
	free(pm);
	return 0;
}


/* int pmdbAdd(char *key,drinkersList *);
 * 	add an element to the database
 * 	return 1 if already exists, 0 on success
 */
int pmdbAdd(pmdb * pm,char *key,drinkersList * dl){
	unsigned short h;
	drinkersList * curr;

	assert(pm);
	assert(pm->data);
	assert(dl);
	pthread_mutex_lock(pm->lock);
	h = sha_hash(key,strlen(key));
	curr = pm->data[h];
	while(curr){
		if(!strcmp(curr->key,key)){
			pthread_mutex_unlock(pm->lock);
			return 1;	// already exists
		}
		curr=curr->next;
	}

	// now that we know it's not already there, add it to the front
	dl->next = pm->data[h];
	pm->data[h] = dl;

	pthread_mutex_unlock(pm->lock);
	return 0;
}

/* drinkersList * pmdbFind(pmdb*, char *key);
 * 	return an entry matching key if it exists
 */
drinkersList * pmdbFind(pmdb*pm, char *key){
	unsigned short h;
	drinkersList * curr;

	assert(pm);
	assert(pm->data);
	assert(key);
	pthread_mutex_lock(pm->lock);
	h = sha_hash(key,strlen(key));
	curr = pm->data[h];
	while(curr){
		if(!strcmp(curr->key,key)){
			pthread_mutex_unlock(pm->lock);
			return curr;	// found it
		}
		curr=curr->next;
	}

	// not found
	pthread_mutex_unlock(pm->lock);
	return NULL;
}

/* do tests
 */

int do_pmdb_tests(){
	pmdb * pm;
	drinkersList *dl1;
	ip_port_t * ip;

	printf("-----------  Doing PMDB tests\n");
	pm = pmdbNew();

	dl1 = drinkersListNew("test1",3);
	ip = ip_port_create("111.111.111.111",1);
	assert(!drinkersListAdd(dl1,ip));
	ip = ip_port_create("222.222.222.222",1);
	assert(!drinkersListAdd(dl1,ip));
	ip = ip_port_create("333.333.333.333",1);
	assert(!drinkersListAdd(dl1,ip));
	ip = ip_port_create("444.444.444.444",1);
	assert(!drinkersListAdd(dl1,ip));
	
	assert(!pmdbAdd(pm,"test1",dl1));
	assert(pmdbAdd(pm,"test1",dl1));
	assert(pmdbFind(pm,"test2")==NULL);
	assert(pmdbFind(pm,"test1")==dl1);
	
	return 0;
}








